""" Test code-links directive """

from os.path import isfile, join as pjoin

import re

from nb2plots.converters import to_pxml

from nb2plots.testing import PlotsBuilder


def test_codelinks():

    def as_pxml(rst_text):
        return to_pxml.from_rst(rst_text, resolve=False)

    page = """\
Text here

.. code-links::

More text here."""
    both_re = re.compile("""\
<document source=".*?">
    <paragraph>
        Text here
    <code_links>
        <bullet_list bullet="\*">
            <list_item>
                <paragraph>
                    <runrole_reference refdoc="contents" reftarget="/contents.py" reftype="pyfile">
                        Download this page as a Python code file
                    ;
            <list_item>
                <paragraph>
                    <runrole_reference refdoc="contents" reftarget="/contents.ipynb" reftype="clearnotebook">
                        Download this page as a Jupyter notebook \(no outputs\)
                    ;
            <list_item>
                <paragraph>
                    <runrole_reference refdoc="contents" reftarget="/contents_full.ipynb" reftype="fullnotebook">
                        Download this page as a Jupyter notebook \(with outputs\)
                    .
    <paragraph>
        More text here.""")
    pxml = as_pxml(page)
    assert both_re.match(pxml)
    # Default is 'both'
    page = """\
Text here

.. code-links:: python clear full

More text here."""
    pxml = as_pxml(page)
    assert both_re.match(pxml)
    page = """\
Text here

.. code-links:: clear

More text here."""
    pxml = as_pxml(page)
    assert re.match("""\
<document source=".*?">
    <paragraph>
        Text here
    <code_links>
        <bullet_list bullet="\*">
            <list_item>
                <paragraph>
                    <runrole_reference refdoc="contents" reftarget="/contents.ipynb" reftype="clearnotebook">
                        Download this page as a Jupyter notebook \(no outputs\)
                    .
    <paragraph>
        More text here.""" , pxml)
    page = """\
Text here

.. code-links:: full

More text here."""
    pxml = as_pxml(page)
    assert re.match("""\
<document source=".*?">
    <paragraph>
        Text here
    <code_links>
        <bullet_list bullet="\*">
            <list_item>
                <paragraph>
                    <runrole_reference refdoc="contents" reftarget="/contents_full.ipynb" reftype="fullnotebook">
                        Download this page as a Jupyter notebook \(with outputs\)
                    .
    <paragraph>
        More text here.""", pxml)
    page = """\
Text here

.. code-links:: full python

More text here."""
    pxml = as_pxml(page)
    assert re.match("""\
<document source=".*?">
    <paragraph>
        Text here
    <code_links>
        <bullet_list bullet="\*">
            <list_item>
                <paragraph>
                    <runrole_reference refdoc="contents" reftarget="/contents_full.ipynb" reftype="fullnotebook">
                        Download this page as a Jupyter notebook \(with outputs\)
                    ;
            <list_item>
                <paragraph>
                    <runrole_reference refdoc="contents" reftarget="/contents.py" reftype="pyfile">
                        Download this page as a Python code file
                    .
    <paragraph>
        More text here.""", pxml)


class TestSubdirCodeLinks(PlotsBuilder):
    """ Test output file locations for code-links directive.
    """

    rst_sources = {'foo/a_page': """\
A section
#########

.. code-links::

More text.
"""}

    def test_output(self):
        for suffix in ('.py', '.ipynb', '_full.ipynb'):
            assert isfile(pjoin(self.out_dir, 'foo', 'a_page' + suffix))



""" Utilities for testing conversion between formats
"""
import difflib

from docutils.core import publish_string
from docutils.writers.pseudoxml import Writer as PXMLWriter
from docutils.io import StringOutput

# Translate inserted smartquote characters back to their original form
_UNSMART_IN = u'\u2019\u201c\u201d\u2026'
_UNSMART_OUT = (u"'", u'"', u'"', u"...")
UNSMART_TABLE = {ord(x): y for x, y in zip(_UNSMART_IN, _UNSMART_OUT)}
_UNSMART_OUT_NB = (u"'", u'\\"', u'\\"', u"...")
UNSMART_TABLE_NB = {ord(x): y for x, y in zip(_UNSMART_IN, _UNSMART_OUT_NB)}


def fcontents(fname, mode='b'):
    with open(fname, 'r' + mode) as fobj:
        contents = fobj.read()
    return contents


def unsmart(in_str):
    # Remove effect of smart quotes
    # See: https://github.com/sphinx-doc/sphinx/issues/3967
    return in_str.translate(UNSMART_TABLE)


def unsmart_nb(in_str):
    # Remove effect of smart quotes in notebook cells
    return in_str.translate(UNSMART_TABLE_NB)


def unsmart_converter(converter, table=None):
    # Decorate function to remove effect of smart quotes.

    def unsmarted(rst_str):
        return unsmart(converter(rst_str))

    return unsmarted


def _diff_strs(first, second):
    # Replicate some of the standard string comparison error message.
    # This is from unittest.TestCase.assertMultiLineEqual
    firstlines = first.splitlines(True)
    secondlines = second.splitlines(True)
    if len(firstlines) == 1 and first.strip('\r\n') == first:
        firstlines = [first + '\n']
        secondlines = [second + '\n']
    return ''.join(difflib.ndiff(firstlines, secondlines))


def convert_assert(rst_str, converter, expected, encoding='utf8'):
    actual = converter(rst_str)
    if (actual == expected):
        assert actual == expected
        return
    # Make some useful debugging output
    if encoding is not None:
        actual = actual.decode(encoding)
        expected = expected.decode(encoding)
    msg = ('actual, expected not equal:\n' +
           _diff_strs(actual, expected))
    pxml = publish_string(rst_str, writer=PXMLWriter())
    msg += '\nwith doctree\n' + pxml.decode(
        'utf8' if encoding is None else encoding)
    assert actual == expected, msg


def doctree_assert(doctree, writer, expected):
    destination = StringOutput(encoding='utf8')
    assert writer.write(doctree, destination) == expected



# ## A section
#
# A test for links.  An [inline link](http://nipy.org).
#
# A [link reference](http://python.org).
#
# An internal reference.
#
# ### Another section
#
# A reference to an [inline link](http://nipy.org).



# Start text.
#
# first definition
#
# :   Here I define the first thing.
#
# second definition
#
# :   Now I define the second thing.  The definition can get long enough that it
#     goes over more than one line.
#
# More text.



""" Package providing utilities for testing
"""

from os.path import abspath, dirname, join as pjoin

import numpy as np

from sphinxtesters import SourcesBuilder


DATA_PATH = abspath(pjoin(
    dirname(__file__),
    '..',
    'tests',
    'rst_md_files'))


def setup_test():
    """ Set numpy print options to "legacy" for new versions of numpy
    """
    from distutils.version import LooseVersion

    if LooseVersion(np.__version__) >= LooseVersion('1.14'):
        np.set_printoptions(legacy="1.13")


class PlotsBuilder(SourcesBuilder):
    """ Class to build pages with nbplots default extensions

    Used by several test functions.
    """

    conf_source = 'extensions = ["nb2plots", "sphinx.ext.doctest"]'



""" Support for command-line scripts
"""

import sys
from argparse import ArgumentParser

from .converters import NbConverter


def bin_stdout():
    """ Bytes stream for writing to stdout """
    return sys.stdout.buffer if sys.version_info[0] > 2 else sys.stdout


def get_parser(description):
    """ Get parser for sphinx2something utilities
    """
    parser = ArgumentParser(description=description)
    parser.add_argument('rst_file', help = 'ReST file to convert')
    parser.add_argument('-W', '--warn-is-error', action='store_true',
                        help = 'turn warnings into errors')
    return parser


def do_main(description, buildername):
    """ Get main clause for sphinx2something utilities
    """
    parser = get_parser(description)
    args = parser.parse_args()
    with open(args.rst_file, 'rt') as fobj:
        contents = fobj.read()
    converter = NbConverter(buildername,
                            status=sys.stderr,
                            warningiserror=args.warn_is_error)
    output = converter.from_rst(contents)
    bin_stdout().write(output.encode('utf-8'))



""" Directive to insert links to current document as runnable code.

As a bare directive, with no options, inserts links to the current document as
a Python code file; a clear notebook (no outputs) and a full notebook (outputs
inserted by executing the built notebook)::

    .. code-links:

You can select one or more of these links with a list as an argument to the
directive, where "python", "clear" and "full" refer to a Python code file,
clear notebook file and a full notebook file, respectively. For example::

    .. code-links: python

    .. code-links: python full

    .. code-links: clear full

``python clear full`` is the default.

Kernel timeout
--------------

When you build the full notebook, Jupyter will execute the code in each cell.
By default, Jupyter will time out of any cell takes longer than 30 seconds to
execute.  You can change this default for the whole project with the
``fill_notebook_timeout`` setting in the ``conf.py`` file (see below).  If you
just want to change the setting for a single page, you can add the ``timeout``
option to the ``code-links`` directive.  For example::

    .. code-links:
        :timeout: 120

Set the timeout value to -1 or ``none`` to disable timeout entirely for this
directive / page.

Configuration options
---------------------

The code-links directive has the following configuration options, that can be
set in the ``conf.py`` file.

    fill_notebook_timeout
        Default value for Jupyter kernel timeout when executing notebooks
        during page build.  If not set, default is 30 seconds. Set to -1 or
        None to disable timeout.
"""

from docutils.statemachine import StringList
from docutils import nodes
from docutils.parsers.rst import Directive

from .runroles import NAME2ROLE, convert_timeout, runrole_reference


class code_links(nodes.container):
    """ Node to contain code links """


class CodeLinks(Directive):
    """ Add links to built runnable code from contained page.
    """

    has_content = False
    required_arguments = 0
    optional_arguments = 3
    option_spec = {'timeout': convert_timeout}

    _type2params = dict(python=dict(role_name='pyfile', suffix=''),
                        clear=dict(role_name='clearnotebook', suffix=''),
                        full=dict(role_name='fullnotebook', suffix='_full'))

    code_links_node = code_links

    def _role_lines(self):
        """ Text lines for notebook roles of types requested in options.
        """
        code_types = (self.arguments if len(self.arguments)
                      else ['python', 'clear', 'full'])
        env = self.state.document.settings.env
        lines = []
        for code_type in code_types:
            params = self._type2params[code_type]
            suffix = params['suffix']
            role_name = params['role_name']
            role = NAME2ROLE[role_name]
            lines.append(
                '* :{role_name}:`{descr} </{docname}{suffix}{ext}>`'.format(
                    role_name=role_name,
                    descr=role.default_text,
                    docname=env.docname,
                    suffix=suffix,
                    ext=role.default_extension))
        return lines

    def run(self):
        lines = []
        body_lines = self._role_lines()
        for i, line in enumerate(body_lines):
            suffix = '.' if i == len(body_lines) - 1 else ';'
            lines.append('{}{}'.format(line, suffix))
        node = self.code_links_node('\n'.join(lines))
        self.state.nested_parse(StringList(lines),
                                self.content_offset,
                                node)
        if 'timeout' in self.options:
            for node in node.traverse(runrole_reference):
                if node['reftype'] == 'fullnotebook':
                    node['timeout'] = self.options['timeout']
        return [node]


def setup(app):
    app.add_directive('code-links', CodeLinks)

    # Pass through containers used as markers for code links
    null = lambda self, node : None
    app.add_node(CodeLinks.code_links_node,
                 **{builder: (null, null)
                    for builder in ('html', 'latex', 'text', 'texinfo')})
    app.add_config_value('fill_notebook_timeout', 30, True)



""" Test mpl-interactive directive """

import re

from nb2plots.converters import to_pxml


def test_mpl_interactive():
    page = """\
Text here

.. mpl-interactive::

More text here."""
    pxml = to_pxml.from_rst(page)
    assert re.match("""\
<document source=".*?">
    <paragraph>
        Text here
    <mpl_hint>
        <paragraph>
            If running in the IPython console, consider running 
            <literal>
                %matplotlib
             to enable
            interactive plots.  If running in the Jupyter Notebook, use 
            <literal>
                %matplotlib
                inline
            .
    <paragraph>
        More text here.""", pxml)
    page = """\
Text here

.. mpl-interactive::

    Any ReST you *like*.

More text here."""
    pxml = to_pxml.from_rst(page)
    assert re.match("""\
<document source=".*?">
    <paragraph>
        Text here
    <mpl_hint>
        <paragraph>
            Any ReST you 
            <emphasis>
                like
            .
    <paragraph>
        More text here.""", pxml)



# -*- coding: utf-8 -*-
"""Simple Markdown writer for reStructuredText.

See:

* https://daringfireball.net/projects/markdown/syntax
* https://guides.github.com/pdfs/markdown-cheatsheet-online.pdf
"""
from __future__ import unicode_literals

import re
from textwrap import dedent
import posixpath

__docformat__ = 'reStructuredText'

from docutils import frontend, nodes, writers, languages
from collections import OrderedDict

from .nbplots import checked_visit


class IndentLevel(object):
    """ Class to hold text being written for a certain indentation level

    For example, all text in list_elements need to be indented.  A list_element
    creates one of these indentation levels, and all text contained in the
    list_element gets written to this IndentLevel.  When we leave the
    list_element, we ``write`` the text with suitable prefixes to the next
    level down, which might be the base of the document (document body) or
    another indentation level, if this is - for example - a nested list.

    In most respects, IndentLevel behaves like a list.
    """
    def __init__(self, base, prefix, first_prefix=None):
        self.base = base  # The list to which we eventually write
        self.prefix = prefix  # Text prepended to lines
        # Text prepended to first list
        self.first_prefix = prefix if first_prefix is None else first_prefix
        # Our own list to which we append before doing a ``write``
        self.content = []

    def append(self, new):
        self.content.append(new)

    def __getitem__(self, index):
        return self.content[index]

    def __len__(self):
        return len(self.content)

    def __bool__(self):
        return len(self) != 0

    def write(self):
        """ Add ``self.contents`` with current ``prefix`` and ``first_prefix``

        Add processed ``self.contents`` to ``self.base``.  The first line has
        ``first_prefix`` prepended, further lines have ``prefix`` prepended.

        Empty (all whitespace) lines get written as bare carriage returns, to
        avoid ugly extra whitespace.
        """
        string = ''.join(self.content)
        lines = string.splitlines(True)
        if len(lines) == 0:
            return
        texts = [self.first_prefix + lines[0]]
        for line in lines[1:]:
            if line.strip() == '':  # avoid prefix for empty lines
                texts.append('\n')
            else:
                texts.append(self.prefix + line)
        self.base.append(''.join(texts))


def _make_method(to_add):
    """ Make a method that adds `to_add`

    We need this function so that `to_add` is a fresh and unique variable at
    the time the method is defined.
    """

    def method(self, node):
        self.add(to_add)

    return method


def add_pref_suff(pref_suff_map):
    """ Decorator adds visit, depart methods for prefix/suffix pairs
    """
    def dec(cls):
        # Need _make_method to ensure new variable picked up for each iteration
        # of the loop.  The defined method picks up this new variable in its
        # scope.
        for key, (prefix, suffix) in pref_suff_map.items():
            setattr(cls, 'visit_' + key, _make_method(prefix))
            setattr(cls, 'depart_' + key, _make_method(suffix))
        return cls

    return dec


def add_pass_thru(pass_thrus):
    """ Decorator adds explicit pass-through visit and depart methods
    """
    def meth(self, node):
        pass

    def dec(cls):
        for element_name in pass_thrus:
            for meth_prefix in ('visit_', 'depart_'):
                meth_name = meth_prefix + element_name
                if hasattr(cls, meth_name):
                    raise ValueError('method name {} already defined'
                                     .format(meth_name))
                setattr(cls, meth_name, meth)
        return cls

    return dec


# Characters that should be escaped in Markdown
ESCAPE_RE = re.compile(r'([\\*`])')

# Doctree elements for which Markdown element is <prefix><content><suffix>
PREF_SUFF_ELEMENTS = {
    'emphasis': ('*', '*'),   # Could also use ('_', '_')
    'strong' : ('**', '**'),  # Could also use ('__', '__')
    'literal' : ('`', '`'),
    'subscript' : ('<sub>', '</sub>'),
    'superscript' : ('<sup>', '</sup>'),
}

# Doctree elements explicitly passed through without extra markup
PASS_THRU_ELEMENTS = ('document',
                      'container',
                      'target',
                      'inline',
                      'definition_list',
                      'definition_list_item',
                      'term',
                      'field_list',
                      'field_list_item',
                      'field',
                      'field_name',
                      'mpl_hint',
                      'pending_xref',
                      'compound',
                     )


@add_pass_thru(PASS_THRU_ELEMENTS)
@add_pref_suff(PREF_SUFF_ELEMENTS)
class Translator(nodes.NodeVisitor):

    std_indent = '    '

    def __init__(self, document, builder=None):
        nodes.NodeVisitor.__init__(self, document)
        self.builder = builder
        self.settings = settings = document.settings
        lcode = settings.language_code
        self.language = languages.get_language(lcode, document.reporter)
        # Not-None here indicates Markdown should use HTTP for internal and
        # download links.
        self.markdown_http_base = (builder.markdown_http_base if builder
                                   else None)
        # Warn only once per writer about unsupported elements
        self._warned = set()
        # Lookup table to get section list from name
        self._lists = OrderedDict((('head', []),
                                   ('body', []),
                                   ('foot', [])))
        # Reset attributes modified by reading
        self.reset()
        # Attribute shortcuts
        self.head, self.body, self.foot = self._lists.values()

    def reset(self):
        """ Initialize object for fresh read """
        for part in self._lists.values():
            part[:] = []

        # Current section heading level during writing
        self.section_level = 0

        # FIFO list of list prefixes, while writing nested lists.  Each element
        # corresponds to one level of nesting.  Thus ['1. ', '1. ', '* '] would
        # occur when writing items of an unordered list, that is nested within
        # an ordered list, that in turn is nested in another ordered list.
        self.list_prefixes = []

        # FIFO list of indentation levels.  When we are writing a block of text
        # that should be indented, we create a new indentation level.  We only
        # write the text when we leave the indentation level, so we can insert
        # the correct prefix for every line.
        self.indent_levels = []

        # Flag indicating we are processing docinfo items
        self._in_docinfo = False

        # Flag for whether to escape characters
        self._escape_text = True

    def astext(self):
        """Return the final formatted document as a string."""
        parts = [''.join(lines).strip() for lines in self._lists.values()]
        parts = [part + '\n\n' for part in parts if part]
        return ''.join(parts).strip() + '\n'

    def ensure_eol(self):
        """Ensure the last line in current base is terminated by new line."""
        out = self.get_current_output()
        if out and out[-1] and out[-1][-1] != '\n':
            out.append('\n')

    def get_current_output(self, section='body'):
        """ Get list or IndentLevel to which we are currently writing """
        return (self.indent_levels[-1] if self.indent_levels
                else self._lists[section])

    def add(self, string, section='body'):
        """ Add `string` to `section` or current output

        Parameters
        ----------
        string : str
            String to add to output document
        section : {'body', 'head', 'foot'}, optional
            Section of document that generated text should be appended to, if
            not already appending to an indent level.  If already appending to
            an indent level, method will ignore `section`.  Use
            :meth:`add_section` to append to a section unconditionally.
        """
        self.get_current_output(section).append(string)

    def add_section(self, string, section='body'):
        """ Add `string` to `section` regardless of current output

        Can be useful when forcing write to header or footer.

        Parameters
        ----------
        string : str
            String to add to output document
        section : {'body', 'head', 'foot'}, optional
            Section of document that generated text should be appended to.
        """
        self._lists[section].append(string)

    def start_level(self, prefix, first_prefix=None, section='body'):
        """ Create a new IndentLevel with `prefix` and `first_prefix`
        """
        base = (self.indent_levels[-1].content if self.indent_levels else
                self._lists[section])
        level = IndentLevel(base, prefix, first_prefix)
        self.indent_levels.append(level)

    def finish_level(self):
        """ Remove most recent IndentLevel and write contents
        """
        level = self.indent_levels.pop()
        level.write()

    def escape_chars(self, txt):
        # Escape (some) characters with special meaning for Markdown
        return ESCAPE_RE.sub(r'\\\1', txt)

    def visit_Text(self, node):
        text = node.astext()
        if self._escape_text:
            text = self.escape_chars(text)
        self.add(text)

    def depart_Text(self, node):
        pass

    def visit_comment(self, node):
        txt = node.astext()
        if txt.strip():
            self.add('<!-- ' + node.astext() + ' -->\n')
        raise nodes.SkipNode

    def visit_docinfo(self, node):
        self._in_docinfo = True

    def depart_docinfo(self, node):
        self._in_docinfo = False

    def process_docinfo_item(self, node):
        """ Called explicitly from methods in this class
        """
        self.add_section('% {}\n'.format(node.astext()), section='head')
        raise nodes.SkipNode

    def visit_definition(self, node):
        self.add('\n\n')
        self.start_level('    ', ':   ')

    def depart_definition(self, node):
        self.finish_level()

    visit_field_body = visit_definition

    depart_field_body = depart_definition

    def visit_paragraph(self, node):
        pass

    def depart_paragraph(self, node):
        self.ensure_eol()
        self.add('\n')

    def visit_math_block(self, node):
        # docutils math block
        self._escape_text = False
        self.add('$$\n')

    def depart_math_block(self, node):
        self._escape_text = True
        self.ensure_eol()
        self.add('$$\n\n')

    def visit_displaymath(self, node):
        # sphinx math blocks become displaymath
        self.add('$$\n{}\n$$\n\n'.format(node['latex']))
        raise nodes.SkipNode

    def visit_math(self, node):
        # sphinx math node has 'latex' attribute, docutils does not
        if 'latex' in node:  # sphinx math node
            self.add('${}$'.format(node['latex']))
            raise nodes.SkipNode
        # docutils math node
        self._escape_text = False
        self.add('$')

    def depart_math(self, node):
        # sphinx node skipped in visit, only docutils gets here
        self._escape_text = True
        self.add('$')

    def visit_literal_block(self, node):
        self._escape_text = False
        code_type = node['classes'][1] if 'code' in node['classes'] else ''
        self.add('```' + code_type + '\n')

    def depart_literal_block(self, node):
        self._escape_text = True
        self.ensure_eol()
        self.add('```\n\n')

    def visit_doctest_block(self, node):
        self._escape_text = False
        self.add('```python\n')

    depart_doctest_block = depart_literal_block

    def visit_block_quote(self, node):
        self.start_level('> ')

    def depart_block_quote(self, node):
        self.finish_level()

    def visit_section(self, node):
        self.section_level += 1

    def depart_section(self, node):
        self.section_level -= 1

    def visit_enumerated_list(self, node):
        self.list_prefixes.append('1. ')

    def depart_enumerated_list(self, node):
        self.list_prefixes.pop()

    def visit_bullet_list(self, node):
        self.list_prefixes.append('* ')

    depart_bullet_list = depart_enumerated_list

    def visit_list_item(self, node):
        first_prefix = self.list_prefixes[-1]
        prefix = ' ' * len(first_prefix)
        self.start_level(prefix, first_prefix)

    def depart_list_item(self, node):
        self.finish_level()

    def visit_problematic(self, node):
        self.add('\n\n```\n{}\n```\n\n'.format(node.astext()))
        raise nodes.SkipNode

    def visit_system_message(self, node):
        if node['level'] < self.document.reporter.report_level:
            # Level is too low to display
            raise nodes.SkipNode
        line = ', line %s' % node['line'] if node.hasattr('line') else ''
        self.add("```\nSystem Message: {}:{}\n\n{}\n```\n\n".format(
            node['source'], line, node.astext()))
        raise nodes.SkipNode

    def visit_title(self, node):
        self.add((self.section_level + 1) * '#' + ' ')

    def depart_title(self, node):
        self.ensure_eol()
        self.add('\n')

    def visit_subtitle(self, node):
        self.add((self.section_level + 2) * '#' + ' ')

    depart_subtitle = depart_title

    def visit_transition(self, node):
        # Simply replace a transition by a horizontal rule.
        # Could use three or more '*', '_' or '-'.
        self.add('\n---\n\n')
        raise nodes.SkipNode

    def _refuri2http(self, node):
        # Replace 'refuri' in reference with HTTP address, if possible
        # None for no possible address
        url = node.get('refuri')
        if not node.get('internal'):
            return url
        # If HTTP page build URL known, make link relative to that.
        if not self.markdown_http_base:
            return None
        this_doc = self.builder.current_docname
        if url in (None, ''):  # Reference to this doc
            url = self.builder.get_target_uri(this_doc)
        else:  # URL is relative to the current docname.
            this_dir = posixpath.dirname(this_doc)
            if this_dir:
                url = posixpath.normpath('{}/{}'.format(this_dir, url))
        url = '{}/{}'.format(self.markdown_http_base, url)
        if 'refid' in node:
            url += '#' + node['refid']
        return url

    def visit_reference(self, node):
        # If no target possible, pass through.
        url = self._refuri2http(node)
        if url is None:
            return
        self.add('[{0}]({1})'.format(node.astext(), url))
        raise nodes.SkipNode

    def depart_reference(self, node):
        pass

    def visit_nbplot_epilogue(self, node):
        raise nodes.SkipNode

    def visit_nbplot_not_rendered(self, node):
        raise nodes.SkipNode

    def visit_code_links(self, node):
        raise nodes.SkipNode

    def visit_index(self, node):
        # Drop index entries
        raise nodes.SkipNode

    def visit_substitution_definition(self, node):
        # Drop substitution definitions - the doctree already contains the text
        # with substitutions applied.
        raise nodes.SkipNode

    def visit_only(self, node):
        if node['expr'] == 'markdown':
            self.add(dedent(node.astext()) + '\n')
        raise nodes.SkipNode

    def visit_runrole_reference(self, node):
        raise nodes.SkipNode

    def visit_download_reference(self, node):
        # If not resolving internal links, or there is no filename specified,
        # pass through.
        filename = node.get('filename')
        if None in (self.markdown_http_base, filename):
            return
        target_url = '{}/_downloads/{}'.format(self.markdown_http_base,
                                               filename)
        self.add('[{}]({})'.format(node.astext(), target_url))
        raise nodes.SkipNode

    def depart_download_reference(self, node):
        pass

    visit_compact_paragraph = visit_paragraph

    depart_compact_paragraph = depart_paragraph

    def visit_nbplot_container(self, node):
        checked_visit(self, node)

    def depart_nbplot_container(self, node):
        pass

    def unknown_visit(self, node):
        """ Warn once per instance for unsupported nodes

        Intercept docinfo items if in docinfo block.
        """
        if self._in_docinfo:
            self.process_docinfo_item(node)
            return
        # We really don't know this node type, warn once per node type
        node_type = node.__class__.__name__
        if node_type not in self._warned:
            self.document.reporter.warning('The ' + node_type +
                ' element not yet supported in Markdown.')
            self._warned.add(node_type)
        raise nodes.SkipNode


class Writer(writers.Writer):

    supported = ('markdown',)
    """Formats this writer supports."""

    output = None
    """Final translated form of `document`."""

    # Add configuration settings for additional Markdown flavours here.
    settings_spec = (
        'Markdown-Specific Options',
        None,
        (('Extended Markdown syntax.',
          ['--extended-markdown'],
          {'default': 0, 'action': 'store_true',
           'validator': frontend.validate_boolean}),
         ('Strict Markdown syntax. Default: true',
          ['--strict-markdown'],
          {'default': 1, 'action': 'store_true',
           'validator': frontend.validate_boolean}),))

    translator_class = Translator

    def __init__(self, builder=None):
        writers.Writer.__init__(self)
        self.builder = builder

    def translate(self):
        visitor = self.translator_class(self.document, self.builder)
        self.document.walkabout(visitor)
        self.output = visitor.astext()



""" Tests for runroles module
"""
import re
from os.path import isfile, join as pjoin

from nb2plots import runroles as rr
from nb2plots.runroles import convert_timeout
from nb2plots import doctree2nb
from nb2plots import doctree2py
from nb2plots.converters import to_pxml

from nb2plots.testing import mockapp
from nb2plots.testing import PlotsBuilder

import pytest


def test_runroles_setup(*args):
    # Test extension setup works as expected
    app = mockapp.get_app()
    rr.setup(app)
    connects = [('doctree-resolved', rr.collect_runfiles),
                ('build-finished', rr.write_runfiles)]
    roles = list(rr.NAME2ROLE.items())
    translators = [('ipynb', doctree2nb.Translator),
                   ('pyfile', doctree2py.Translator)]
    for method_name, args, kwargs in app.method_calls:
        if (method_name == 'connect' and args[0:2] in connects):
            connects.remove(args[0:2])
        if (method_name == 'add_role' and args[0:2] in roles):
            roles.remove(args[0:2])
        if (method_name == 'set_translator' and args[0:2] in translators):
            translators.remove(args[0:2])

    assert len(connects) == 0, 'Connections failed'
    assert len(roles) == 0, 'Roles failed'
    assert len(translators) == 0, 'Translators failed'


def test_convert_timeout():
    assert convert_timeout('4') == 4
    assert convert_timeout('30') == 30
    assert convert_timeout('-1') == -1
    assert convert_timeout('None') == None
    assert convert_timeout('none') == None
    assert convert_timeout('noNe') == None
    with pytest.raises(ValueError):
        convert_timeout('-2')


def test_runrole_doctrees():
    # Test that run roles generate expected doctrees
    expected_re_fmt = """\
<document source=".*?">
    <paragraph>
        Text then 
        <runrole_reference filename="{filebase}.{ext}" refdoc="contents" reftarget="{base}.{ext}" reftype="{role_type}">
            {descr}
         then text."""

    def assert_rst_pxml(pxml_params, rst_source):
        code_type = pxml_params['code_type']
        pxml = to_pxml.from_rst(rst_source)
        if not 'ext' in pxml_params:
            pxml_params['ext'] = 'py' if code_type == 'pyfile' else 'ipynb'
        pxml_regex = expected_re_fmt.format(
            role_type=code_type,
            **pxml_params)
        assert re.match(pxml_regex, pxml)

    assert_rst_pxml(
        dict(code_type='clearnotebook',
             filebase='contents',
             base='/contents',
             descr='Download this page as a Jupyter notebook \(no outputs\)'),
        "Text then :clearnotebook:`.` then text.")
    assert_rst_pxml(
        dict(code_type='fullnotebook',
             filebase='contents',
             base='/contents',
             descr=('Download this page as a Jupyter notebook '
                    '\(with outputs\)')),
        "Text then :fullnotebook:`.` then text.")
    assert_rst_pxml(
        dict(code_type='pyfile',
             filebase='contents',
             base='/contents',
             descr='Download this page as a Python code file'),
        "Text then :pyfile:`.` then text.")
    for code_type in ('clearnotebook', 'fullnotebook', 'pyfile'):
        assert_rst_pxml(
            dict(code_type=code_type,
                 filebase='contents',
                 base='/contents',
                 descr='message to taste'),
            "Text then :{}:`message to taste` then text.".format(code_type))
        assert_rst_pxml(
            dict(code_type=code_type,
                 filebase='foo',
                 base='foo',
                 ext='ipynb',
                 descr='message to taste'),
            "Text then :{}:`message to taste <foo.ipynb>` then text."
            .format(code_type))
        # It is annoying, but there must be a first text description for the
        # angle brackets part to refer to the output file name.  This is how
        # ReST role processing usually works.
        assert_rst_pxml(
            dict(code_type=code_type,
                 filebase='contents',
                 base='/contents',
                 descr='<foo.ipynb>'),
            "Text then :{}:`<foo.ipynb>` then text.".format(code_type))


class TestSubdirBuild(PlotsBuilder):
    """ Test that output files from subdirectories have correct location
    """

    rst_sources = {'foo/a_page': """\
A section
#########

Some text.

:clearnotebook:`notebook here`

Text is endless.

:pyfile:`code here`

Bare path is relative to containing directory:

:pyfile:`code here <my_code.py>`

Can have relative parts too:

:clearnotebook:`notebook here <../my_nb.ipynb>`

Prepended / refers to root of project:

:pyfile:`code here </more_code.py>`
"""}

    def test_output(self):
        for suffix in ('.py', '.ipynb'):
            assert isfile(pjoin(self.out_dir, 'foo', 'a_page' + suffix))
        assert isfile(pjoin(self.out_dir, 'foo', 'my_code.py'))
        assert isfile(pjoin(self.out_dir, 'my_nb.ipynb'))
        assert isfile(pjoin(self.out_dir, 'more_code.py'))


class TestPyfileAlias(PlotsBuilder):
    """ Test that 'codefile' alias works for pyfile
    """

    rst_sources = {'a_page': """\
A section
#########

:codefile:`code here`

>>> True
True
"""}

    def test_output(self):
        assert isfile(pjoin(self.out_dir, 'a_page.py'))


class TestDuplicatesOK(PlotsBuilder):
    """ Test that same and different filename for same code type works.
    """

    rst_sources = {'a_page': """\
Title
#####

:fullnotebook:`name <foo.ipynb>`

:fullnotebook:`name <foo.ipynb>`

:fullnotebook:`name <bar.ipynb>`

"""}

    def test_output(self):
        assert isfile(pjoin(self.out_dir, 'foo.ipynb'))
        assert isfile(pjoin(self.out_dir, 'bar.ipynb'))


class TestDuplicatesNotOK(PlotsBuilder):
    """ Test that same filename for different code type fails.
    """

    rst_sources = {'a_page': """\
Title
#####

:fullnotebook:`name <foo.ipynb>`

:clearnotebook:`name <foo.ipynb>`

"""}

    should_error = True



""" Utility to show differences between strings

Used for debugging test output
"""

def get_diff(str1, str2, fmt_indent=None):
    """ Return line-by-line difference between multiline strings """
    output = []
    str_lists = [str1.splitlines(), str2.splitlines()]
    lens = [len(L) for L in str_lists]
    min_len = min(lens)
    fmt_str = 'L{:d}, {:2s}: {}'
    if fmt_indent is None:
        fmt_indent = len(fmt_str.format(0, '', '')) * ' '
    for i in range(min(lens)):
        L1 = str_lists[0][i]
        L2 = str_lists[1][i]
        if L1 == L2:
            output.append(fmt_str.format(i, 'OK', L1))
        else:
            output.append(fmt_str.format(i, 'S1', L1))
            output.append(fmt_str.format(i, 'S2', L2))
    diff = lens[1] - lens[0]
    if diff != 0:
        longer = diff > 0
        abs_diff = diff if diff > 0 else -diff
        output.append('Remaining line{} in {} string {}:'.format(
            's' if abs_diff > 1 else '',
            'second' if longer else 'first',
            'are' if abs_diff > 1 else 'is',
        ))
        output += [fmt_indent + L for L in str_lists[longer][min_len:]]
    return '\n'.join(output)



""" Writer to convert doctree to Python .py file
"""
from __future__ import unicode_literals

import re
from textwrap import dedent

from docutils import nodes

from . import doctree2md as d2m


# The following regular expression comes from Python source file "doctest.py".
# License for that file recorded as:
#
# Released to the public domain 16-Jan-2001, by Tim Peters (tim@python.org).
#
# This regular expression is used to find doctest examples in a
# string.  It defines three groups: `source` is the source code
# (including leading indentation and prompts); `indent` is the
# indentation of the first (PS1) line of the source code; and
# `want` is the expected output (including leading indentation).
_EXAMPLE_RE = re.compile(r'''
    # Source consists of a PS1 line followed by zero or more PS2 lines.
    (?P<source>
        (?:^(?P<indent> [ ]*) >>>    .*)    # PS1 line
        (?:\n           [ ]*  \.\.\. .*)*)  # PS2 lines
    \n?
    # Want consists of any non-blank lines that do not start with PS1.
    (?P<want> (?:(?![ ]*$)    # Not a blank line
                 (?![ ]*>>>)  # Not a line starting with PS1
                 .+$\n?       # But any other line
              )*)
    ''', re.MULTILINE | re.VERBOSE)


def parse_doctest(doctest_txt):
    txt = dedent(doctest_txt.expandtabs())
    parts = []
    for m in _EXAMPLE_RE.finditer(txt):
        indent = len(m.group('indent'))
        source_lines = m.group('source').splitlines()
        source = '\n'.join([L[indent + 4:] for L in source_lines])
        parts.append(source)
    return '\n'.join(parts)


class Translator(d2m.Translator):

    def __init__(self, document, builder=None):
        d2m.Translator.__init__(self, document, builder)
        self._in_nbplot = False
        self._init_output()

    def _init_output(self):
        self._out_lines = []

    def reset(self):
        d2m.Translator.reset(self)
        self._in_nbplot = False

    def flush_text(self):
        txt = d2m.Translator.astext(self).strip()
        if txt:
            self._add_text_block(txt)
        self.reset()

    def _add_text_block(self, txt):
        self._out_lines += ['# ' + line if line.strip() else '#'
                            for line in txt.splitlines()]
        self._out_lines.append('')

    def add_code_block(self, txt):
        self.flush_text()
        self._out_lines += txt.splitlines() + ['']

    def astext(self):
        """ Return the document as a string """
        self.flush_text()
        return '\n'.join(self._out_lines)

    def visit_doctest_block(self, node):
        doctest_txt = node.astext().strip()
        if doctest_txt:
            self.add_code_block(parse_doctest(doctest_txt))
        raise nodes.SkipNode

    def visit_nbplot_container(self, node):
        d2m.Translator.visit_nbplot_container(self, node)
        self._in_nbplot = True

    def depart_nbplot_container(self, node):
        self._in_nbplot = False

    def visit_runrole_reference(self, node):
        raise nodes.SkipNode

    def visit_literal_block(self, node):
        """ A literal block may be in an nbplot container """
        if not self._in_nbplot:
            return d2m.Translator.visit_literal_block(self, node)
        self.add_code_block(node.astext())
        raise nodes.SkipNode


class Writer(d2m.Writer):

    supported = ('pyfile',)
    """Formats this writer supports."""

    output = None
    """Final translated form of `document`."""

    translator_class = Translator



# ## Refereed section
#
# This section refers to itself.
#
# It also refers forward to the next section.
#
# Then, and finally, it refers to itself with its own name: Refereed section.
#
# ## Rerefereed
#
# This section refers to this document at Refereed section, and with an
# explicit title, to this document.
#
# Now `contents.rst`.



""" Configuration for py.test test run
"""

def pytest_ignore_collect(path, config):
    """ Skip the origin Gohlke transforms for doctests.

    That file needs some specific doctest setup.
    """
    return path.basename in ('conf.py', 'rst_md_files')



# ## Top section
#
# First paragraph.
#
# Second paragraph.
#
# Third paragraph.
#
# ### H2 section
#
# More text.
#
# ### H2 section again
#
# Continued text.
#
# #### H3 section
#
# Drones on.
#
# ##### H4 section
#
# Sink-hole boring.
#
# #### H3 section again
#
# Still boring.
#
# ### H2 section again
#
# As if from the start.



""" Utils for testing notebooks
"""

from nb2plots.ipython_shim import nbf


def assert_nb_equiv(ipynb, expected):
    actual_nb = nbf.reads(ipynb)
    expected_nb = nbf.reads(expected)
    # Ignore different minor versions of Notebook format
    # It does not appear to be possible to request specific minor versions of
    # the Notebook format.
    expected_nb['nbformat_minor'] = actual_nb['nbformat_minor']
    assert actual_nb == expected_nb



# ## An interesting example
#
# This is an interesting example.

import numpy as np
x = np.linspace(0, 2 * np.pi, 1000)
x[:10]

# Even more interesting than that:
#
# If running in the IPython console, consider running `%matplotlib` to enable
# interactive plots.  If running in the Jupyter Notebook, use `%matplotlib
# inline`.

import matplotlib.pyplot as plt

plt.plot(x, np.sin(x))



""" Tests for build using nbplot extension """
from __future__ import unicode_literals

from os.path import (join as pjoin, dirname, isdir)
import re
import os

from docutils.nodes import paragraph, title

from nb2plots.nbplots import (run_code, parse_parts, nbplot_container,
                              nbplot_epilogue)
from sphinxtesters import SourcesBuilder

from nb2plots.testing import PlotsBuilder

from nb2plots.testing.nbtesters import assert_nb_equiv

import pytest


HERE = dirname(__file__)


def get_otherpage(fname):
    with open(pjoin(HERE, 'otherpages', fname), 'rt') as fobj:
        return fobj.read()


def file_same(file1, file2):
    with open(file1, 'rb') as fobj:
        contents1 = fobj.read()
    with open(file2, 'rb') as fobj:
        contents2 = fobj.read()
    return contents1 == contents2


def test_run_code():
    # Test run_code function
    ns1 = run_code('a = 10')
    assert ns1['a'] == 10
    assert not 'b' in ns1
    # New namespace by default
    ns2 = run_code('b = 20')
    assert ns2['b'] == 20
    assert not 'a' in ns2
    # Adding to a namespace
    ns3 = run_code('c = 30', ns=ns1)
    assert ns3 is ns1
    assert ns1['c'] == 30
    # Checking raises
    run_code('d', raises=NameError)


class TestNbplots(SourcesBuilder):

    conf_source = ('extensions = ["nb2plots"]\n'
                   'nbplot_include_source = False\n')

    rst_sources = dict(a_page=get_otherpage('some_plots.rst'))

    def test_some_plots(self):
        assert isdir(self.out_dir)

        def plot_file(num):
            return pjoin(self.out_dir, 'a_page-{0}.png'.format(num))

        range_10, range_6, range_4 = [plot_file(i) for i in range(1, 4)]
        # Plot 5 is range(6) plot
        assert file_same(range_6, plot_file(5))
        # Plot 7 is range(4) plot
        assert file_same(range_4, plot_file(7))
        # Plot 8 uses the old range(4) figure and the new range(6) figure
        assert file_same(range_4, plot_file('8_00'))
        assert file_same(range_6, plot_file('8_01'))
        # Plot 9 shows the default close-figures behavior in action
        assert file_same(range_4, plot_file(9))
        # Plot 9 does not include source
        html_contents = self.get_built_file('a_page.html')
        # Plot 10 has included source
        assert '# Only a comment' in html_contents


class TestDefaultSource(PlotsBuilder):
    """ Check that default is to include source, not source links """

    rst_sources = dict(a_page="""\
A title
-------

.. nbplot::

    # Only a comment
""")

    def test_include_source_default(self):
        # Plot 1 has included source
        html_contents = self.get_built_file('a_page.html')
        assert '# Only a comment' in html_contents
        # Plot 1 has no source link
        html_contents = self.get_built_file('a_page.html')
        assert 'href=".//a_page-1.py">Source code</a>' not in html_contents


class TestAnnoyingParens(PlotsBuilder):
    """ Test we've fixed the empty parens bug

    The matplotlib plotter puts an annoying empty open/close parens in the
    output when html source link is off, and there are no figures.
    """

    conf_source = ('extensions = ["nb2plots"]\n'
                   'nbplot_html_show_source_link = False')

    rst_sources = dict(a_page="""\
A title
-------

.. nbplot::

    # Only a comment
""")

    def test_annoying_parens(self):
        # Plot 1 has included source
        assert not '<p>()</p>' in self.get_built_file('a_page.html')


class TestDefaultContext(PlotsBuilder):
    """ Test that default context is to keep across plots, reset each doc
    """

    rst_sources = dict(a_page="""\
A title
-------

.. nbplot::

    # The namespace reset at the beginning of each document
    assert 'a' not in globals()
    a = 1

Some text.

.. nbplot::

    b = a
    # A plot preserved across nbplot directives
    plt.plot(range(10))

More text.

.. nbplot::
    :keepfigs:

    # This one should result in the same plot as the previous nbplot
    b = b + 3

Yet more text.

.. nbplot::

    # Here, no plot, without the keepfigs directive
    assert b == 4

""",

                      another_page="""\
Another title
-------------

.. nbplot::

    # The namespace reset at the beginning of each document
    assert 'a' not in globals()
    a = 2

Some text.

.. nbplot::

    c = a

""")

    def test_rebuild_context(self):
        # Does rebuilding still delete context? (Tested in nbplots asserts)
        with open(pjoin(self.page_source, 'another_page.rst'), 'a') as fobj:
            fobj.write('\nSomething added\n')
        with open(pjoin(self.page_source, 'a_page.rst'), 'a') as fobj:
            fobj.write('\nSomething added\n')
        self.__class__.build_source()


class TestRcparams(PlotsBuilder):
    """ Test that rcparams get applied and kept across plots in documents
    """
    conf_source = ('extensions = ["nb2plots"]\n'
                   'nbplot_rcparams = {"text.color": "red"}\n')
    rst_sources = dict(a_page="""\
The start
---------

Plot 1

.. nbplot::

    plt.text(0, 0, "I'm Mr Brightside", color='red')

Plot 2 - shows the default is the same:

.. nbplot::

    plt.text(0, 0, "I'm Mr Brightside")

Plot 3 - changes the default:

.. nbplot::

    plt.rcParams['text.color'] = 'blue'
    plt.text(0, 0, 'Open up my eager eyes')

Plot 4 - new default is blue:

.. nbplot::

    plt.text(0, 0, 'Open up my eager eyes', color='blue')

""",
                       b_page="""
Another title
-------------

Plot color resumes at red:

.. nbplot::

    plt.text(0, 0, "I'm Mr Brightside")

.. nbplot::

    plt.rcParams['text.color'] = 'blue'
    plt.text(0, 0, "Open up my eager eyes")

""")

    def test_rcparams(self):
        # Test plot rcparams applied at beginning of page

        def gpf(name, num):
            # Get plot file
            return pjoin(self.out_dir, '{0}-{1}.png'.format(name, num))

        red_bright = gpf('a_page', 2)
        blue_eager = gpf('a_page', 4)
        assert file_same(gpf('a_page', 1), red_bright)
        assert file_same(gpf('a_page', 3), blue_eager)
        assert file_same(gpf('b_page', 1), red_bright)


class TestDefaultPre(PlotsBuilder):
    """ Check that default pre code is importing numpy as pyplot

    Tested in plot directive body
    """

    rst_sources=dict(a_page="""\
A title
-------

.. nbplot::

    np.inf
    plt.plot(range(10))
""")


class TestNonDefaultPre(PlotsBuilder):
    """ Check that pre code is run in fresh plot context

    Tested in plot directive body
    """
    conf_source=('extensions = ["nb2plots"]\n'
                 'nbplot_pre_code = "import numpy as foo; bar = 1"\n')

    rst_sources=dict(a_page="""\
A title
-------

.. nbplot::

    foo.inf
    assert bar == 1
""")


class TestHiddenDoctests(PlotsBuilder):
    """ Check that doctest code gets hidden but still run

    Build using text builder to get more simply testable output.
    """

    builder = 'text'

    rst_sources=dict(a_page="""\
A title
-------

.. nbplot::
    :include-source: false

    >>> a = 1
    >>> b = 2

Text1

.. nbplot::

    >>> assert a == 1
    >>> assert b == 2

Text2

.. nbplot::
    :include-source: false

    c = 3

Text3

.. nbplot::

    >>> assert 'c' in globals()
""")

    def test_whats_in_the_page(self):
        txt_contents = self.get_built_file('a_page.txt')
        assert 'a = 1' not in txt_contents
        assert 'b = 2' not in txt_contents
        assert 'a == 1' in txt_contents
        assert 'b == 2' in txt_contents
        assert 'c = 3' not in txt_contents


class TestMoreDoctests(PlotsBuilder):
    """ Check that doctest code gets hidden but still tested

    Build using doctest builder
    """

    builder = 'doctest'

    rst_sources=dict(a_page="""\
A title
-------

.. nbplot::
    :include-source: false

    >>> a = 1
    >>> b = 2

Text1

.. nbplot::

    >>> a
    1
    >>> b
    2

Text2

.. nbplot::
    :include-source: false

    c = 3

Text3

.. nbplot::

    >>> 'c' not in globals()
    True
""")


class TestNoRaises(PlotsBuilder):
    """ Confirm that exception, without raises option, generates error
    """
    should_error = True

    rst_sources=dict(a_page="""\
A title
-------

.. nbplot::

    # Another comment
    raise ValueError
""")


class TestRaisesOption(PlotsBuilder):
    """ Check raises option to nbplot directive proceeds without error
    """

    rst_sources=dict(a_page="""\
A title
-------

.. nbplot::
    :raises: ValueError

    # Another comment
    raise ValueError
""")

    def test_include_source_default(self):
        # Check that source still included
        assert '# Another comment' in self.get_built_file('a_page.html')


class TestReference(PlotsBuilder):
    """ Check that reference is correctly generated.

    First check that the code gets built for html as predicted
    """

    builder = 'pseudoxml'

    rst_sources=dict(a_page="""\
A title
-------

.. _a-ref:

.. nbplot::

    >>> a = 1

See :ref:`the ref <a-ref>`.
""")

    def test_reference(self):
        # Check that reference correctly included
        built = self.get_built_file('a_page.pseudoxml')
        expected_regexp = re.compile(
r"""<document _plot_counter="1" source=".*?a_page.rst">
    <section ids="a-title" names="a\\ title">
        <title>
            A title
        <target ids="a-ref" names="a-ref">
        <nbplot_container>
            <doctest_block.*>
                >>> a = 1
        <nbplot_epilogue>
            <.*>
        <paragraph>
            See\s
            <reference internal="True" refid="a-ref">
                <inline.*?>
                    the ref
            \.""", re.DOTALL)
        assert expected_regexp.match(built)


class TestFlags(PlotsBuilder):
    """ Check flags get set correctly
    """

    builder = 'pseudoxml'

    rst_sources=dict(a_page="""\
A title
-------

.. nbplot-flags::

    a = 1
    b = 2

.. nbplot-show-flags::

Some text

.. nbplot-flags::

    c = 3

.. nbplot-show-flags::
""")

    def test_flags(self):
        # Check that flags correctly set from flag directives
        built = self.get_built_file('a_page.pseudoxml')
        assert """
        <title>
            A title
        <literal_block xml:space="preserve">
            {'a': 1, 'b': 2}
        <paragraph>
            Some text
        <literal_block xml:space="preserve">
            {'a': 1, 'b': 2, 'c': 3}""" in built


class TestFlagsConfig(TestFlags):
    """ Check flags set from global config """

    conf_source=("""\
extensions = ["nb2plots"]
nbplot_flags = {'flag1': 5, 'flag2': 6}
""")

    def test_flags(self):
        # Check that global flags merged with local
        built = self.get_built_file('a_page.pseudoxml')
        assert ("""
        <title>
            A title
        <literal_block xml:space="preserve">
            {'a': 1, 'b': 2, 'flag1': 5, 'flag2': 6}
        <paragraph>
            Some text
        <literal_block xml:space="preserve">
            {'a': 1, 'b': 2, 'c': 3, 'flag1': 5, 'flag2': 6}"""
                    in built)


class TestWithoutSkip(PlotsBuilder):
    """ Check that doctest code can be bracketed into skipped, not skipped

    First check that the code gets built for txt as predicted
    """
    builder = 'text'
    conf_source = ('extensions = ["nb2plots", "sphinx.ext.doctest"]\n'
                   'nbplot_flags = {"skip": False}')

    rst_sources=dict(a_page="""\
A title
-------

.. nbplot::

    >>> # always
    >>> a = 'default'

Some text

.. nbplot::
    :render-parts: 1 if skip else 0
    :run-parts: 1 if skip else 0

    >>> a = 'skip is False'

    .. part

    >>> a = 'skip is True'

Keep text coming

.. nbplot::
    :render-parts: 1 if skip else 2
    :run-parts: 1 if skip else 2

    >>> # An empty part, never used

    .. part

    >>> b = 'skip appears to be True'
    >>> a == 'skip is True'
    True

    .. part

    >>> b = 'skip appears to be False'
    >>> a == 'skip is False'
    True

Text continues

.. nbplot::
    :run-parts: 1 if skip else 0

    >>> # doctest only run when skip flag False, always rendered
    >>> b == 'skip appears to be False'
    True

    .. part

    >>> # only when skip flag True
    >>> b == 'skip appears to be True'
    True
""")

    def test_pages(self):
        # Test that the skip=False sections selected
        txt = self.get_built_file('a_page.txt')
        assert ">>> # always\n>>> a = 'default'" in txt
        assert ">>> a = 'skip is False'" in txt
        assert ">>> a = 'skip is True'" not in txt
        # Note ==, distinguishing from test above
        assert ">>> a == 'skip is True'" not in txt
        assert ">>> a == 'skip is False'" in txt
        assert ">>> b == 'skip appears to be False'" in txt
        assert ">>> b == 'skip appears to be True'" not in txt


class TestWithoutSkipDoctest(TestWithoutSkip):
    builder = 'doctest'

    def test_pages(self):
        # No pages built by doctest
        return


class TestWithoutSkipStructure(TestWithoutSkip):
    builder = 'pseudoxml'

    def test_pages(self):
        # Test that the skip=False sections selected
        p_xml = self.get_built_file('a_page.pseudoxml')
        regex = re.compile(
            r"""<document _plot_counter="\d" source=".+?">
    <section ids="a-title" names="a\\ title">
        <title>
            A title
        <nbplot_container>
            <doctest_block xml:space="preserve">
                >>> # always
                >>> a = 'default'
        <nbplot_epilogue>
            <comment xml:space="preserve">
            <comment xml:space="preserve">
            <comment xml:space="preserve">
        <paragraph>
            Some text
        <nbplot_container>
            <doctest_block xml:space="preserve">
                >>> a = 'skip is False'
        <nbplot_epilogue>
            <comment xml:space="preserve">
            <comment xml:space="preserve">
            <comment xml:space="preserve">
        <paragraph>
            Keep text coming
        <nbplot_container>
            <doctest_block xml:space="preserve">
                >>> b = 'skip appears to be False'
                >>> a == 'skip is False'
                True
        <nbplot_epilogue>
            <comment xml:space="preserve">
            <comment xml:space="preserve">
            <comment xml:space="preserve">
        <paragraph>
            Text continues
        <nbplot_container>
            <doctest_block xml:space="preserve">
                >>> # doctest only run when skip flag False, always rendered
                >>> b == 'skip appears to be False'
                True""")
        assert(regex.match(p_xml) is not None)


class TestWithSkip(TestWithoutSkip):
    """ Check that doctest code can be skipped according to flag
    """
    conf_source = ('extensions = ["nb2plots", "sphinx.ext.doctest"]\n'
                   'nbplot_flags = {"skip": True}')

    def test_pages(self):
        # Test that the skip=True sections selected
        txt = self.get_built_file('a_page.txt')
        assert ">>> # always\n>>> a = 'default'" in txt
        assert ">>> a = 'skip is False'" not in txt
        assert ">>> a = 'skip is True'" in txt
        # Note ==, distinguishing from test above
        assert ">>> a == 'skip is True'" in txt
        assert ">>> a == 'skip is False'" not in txt
        # The rendered version always has the first section, regardless of skip
        assert ">>> b == 'skip appears to be False'" in txt
        assert ">>> b == 'skip appears to be True'" not in txt


class TestWithSkipStructure(TestWithSkip):
    builder = 'pseudoxml'

    def test_pages(self):
        # Test that the skip=True sections selected
        p_xml = self.get_built_file('a_page.pseudoxml')
        regex = re.compile(
            r"""<document _plot_counter="\d" source=".+?">
    <section ids="a-title" names="a\\ title">
        <title>
            A title
        <nbplot_container>
            <doctest_block xml:space="preserve">
                >>> # always
                >>> a = 'default'
        <nbplot_epilogue>
            <comment xml:space="preserve">
            <comment xml:space="preserve">
            <comment xml:space="preserve">
        <paragraph>
            Some text
        <nbplot_container>
            <doctest_block xml:space="preserve">
                >>> a = 'skip is True'
        <nbplot_epilogue>
            <comment xml:space="preserve">
            <comment xml:space="preserve">
            <comment xml:space="preserve">
        <paragraph>
            Keep text coming
        <nbplot_container>
            <doctest_block xml:space="preserve">
                >>> b = 'skip appears to be True'
                >>> a == 'skip is True'
                True
        <nbplot_epilogue>
            <comment xml:space="preserve">
            <comment xml:space="preserve">
            <comment xml:space="preserve">
        <paragraph>
            Text continues
        <nbplot_container hide-from="doctest">
            <dont_doctest_block xml:space="preserve">
                >>> # doctest only run when skip flag False, always rendered
                >>> b == 'skip appears to be False'
                True
        <nbplot_epilogue>
            <comment xml:space="preserve">
            <comment xml:space="preserve">
            <comment xml:space="preserve">
        <nbplot_container hide-from="all" show-to="doctest">
            <doctest_block xml:space="preserve">
                >>> # only when skip flag True
                >>> b == 'skip appears to be True'
                True""")
        assert(regex.match(p_xml) is not None)


class TestWithSkipDoctest(TestWithSkip):
    builder = 'doctest'

    def test_pages(self):
        # No pages built by doctest
        return


class TestOtherWD(PlotsBuilder):
    """ Check that it is possible to run code with other working directory
    """

    rst_sources=dict(a_page="""\
A title
-------

.. nbplot::

    >>> # Code run during page generation, e.g. html build
    >>> import os
    >>> assert os.getcwd().endswith('my_wd')
    >>> # Working directory is on Python PATH
    >>> import funky_module
""")

    @classmethod
    def modify_source(cls):
        super(TestOtherWD, cls).modify_source()
        work_dir = pjoin(cls.build_path, 'my_wd')
        os.mkdir(work_dir)
        with open(pjoin(work_dir, 'funky_module.py'), 'wt') as fobj:
            fobj.write('# A funky module\n')
        with open(pjoin(cls.page_source, 'conf.py'), 'at') as fobj:
            fobj.write('\nnbplot_working_directory = "{}"\n'.format(work_dir))


class TestClearNotebook(PlotsBuilder):
    """ Test build of clear notebook with clearnotebook role
    """

    builder = 'text'

    rst_sources=dict(a_page="""\
A title
-------

:clearnotebook:`.`
:clearnotebook:`clear notebook <clear.ipynb>`
:fullnotebook:`full notebook <full.ipynb>`

Some text.

.. nbplot::

    >>> a = 1
    >>> a
    1
""")

    def test_pages(self):
        txt = self.get_built_file('a_page.txt')
        assert re.match(
            r'\n?A title\n\*{7}\n\n\nSome text.\n\n>>> a = 1\n>>> a\n1\n',
            txt)
        ipynb = self.get_built_file('a_page.ipynb')
        assert_nb_equiv(ipynb, r"""
{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## A title\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "Some text."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "a = 1\n",
    "a"
   ]
  }
 ],
 "metadata": {},
 "nbformat": 4,
 "nbformat_minor": 1
}""")
        assert self.get_built_file('clear.ipynb') == ipynb
        full = self.get_built_file('full.ipynb')
        assert_nb_equiv(full, r"""
{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## A title\n",
    "\n",
    "\n",
    "\n",
    "\n",
    "Some text."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/plain": [
       "1"
      ]
     },
     "execution_count": 1,
     "metadata": {},
     "output_type": "execute_result"
    }
   ],
   "source": [
    "a = 1\n",
    "a"
   ]
  }
 ],
 "metadata": {},
 "nbformat": 4,
 "nbformat_minor": 1
}""")


def test_part_finding():
    assert parse_parts([]) == [{'contents': []}]
    assert (parse_parts(['a = 1', 'b = 2']) ==
                             [{'contents': ['a = 1', 'b = 2']}])
    assert (parse_parts(['a = 1', 'b = 2', '', '.. part', '', 'c = 4']) ==
                             [{'contents': ['a = 1', 'b = 2']},
                              {'contents': ['c = 4']}])
    # Need blank lines between
    assert (parse_parts(['a = 1', 'b = 2', '.. part', '', 'c = 4']) ==
            [{'contents': ['a = 1', 'b = 2', '.. part', '', 'c = 4']}])
    assert (parse_parts(['a = 1', 'b = 2', '', '.. part', 'c = 4']) ==
            [{'contents': ['a = 1', 'b = 2', '', '.. part', 'c = 4']}])
    # Add some attributes
    assert (parse_parts(['a = 1', 'b = 2', '',
                         '.. part', ' foo=bar', ' baz=boo',
                         '', 'c = 4']) ==
            [{'contents': ['a = 1', 'b = 2']},
             {'contents': ['c = 4'],
              'foo': 'bar', 'baz': 'boo'}])
    # Can have spaces around the equals
    assert (parse_parts(['a = 1', 'b = 2', '',
                         '.. part', ' foo =bar', ' baz= boo',
                         '', 'c = 4']) ==
            [{'contents': ['a = 1', 'b = 2']},
             {'contents': ['c = 4'],
              'foo': 'bar', 'baz': 'boo'}])
    assert (parse_parts(['a = 1', 'b = 2', '',
                         '.. part', ' foo = bar', ' baz=  boo',
                         '', 'c = 4']) ==
            [{'contents': ['a = 1', 'b = 2']},
             {'contents': ['c = 4'],
              'foo': 'bar', 'baz': 'boo'}])
    # Cannot continue on same line as part separator
    assert (parse_parts(['a = 1', 'b = 2', '',
                         '.. part foo=bar',
                         '', 'c = 4']) ==
            [{'contents': ['a = 1', 'b = 2', '',
                           '.. part foo=bar',
                           '', 'c = 4']}])
    # Must be indentation
    with pytest.raises(ValueError):
        parse_parts(['a = 1', 'b = 2', '',
                     '.. part', 'foo=bar',
                     '', 'c = 4'])
    # Must be same indentation
    with pytest.raises(ValueError):
        parse_parts(['a = 1', 'b = 2', '',
                     '.. part', ' foo=bar', 'baz=boo',
                     '', 'c = 4'])
    with pytest.raises(ValueError):
        parse_parts(['a = 1', 'b = 2', '',
                     '.. part', ' foo=bar', '  baz=boo',
                     '', 'c = 4'])
    # Add some attributes in the first part
    assert (parse_parts(['.. part', ' mr=brightside', ' eager=eyes', '',
                         'a = 1', 'b = 2', '',
                         '.. part', ' foo=bar', ' baz=boo', '', 'c = 4']) ==
            [{'contents': ['a = 1', 'b = 2'],
              'mr': 'brightside',
              'eager': 'eyes'},
             {'contents': ['c = 4'],
              'foo': 'bar', 'baz': 'boo'}])
    # Contents with spaces, leading and trailing spaces skipped
    assert (parse_parts(['a = 1', 'b = 2', '',
                         '.. part', ' foo=[1, 2, s]', ' bar= more stuff ',
                         '', 'c = 4']) ==
            [{'contents': ['a = 1', 'b = 2']},
             {'contents': ['c = 4'],
              'foo': '[1, 2, s]', 'bar': 'more stuff'}])


class TestHideShow(PlotsBuilder):
    """ Check that hide-from and show-to options respected
    """

    conf_source = ('extensions = ["nb2plots", "sphinx.ext.doctest"]\n')

    builder = 'text'

    rst_sources=dict(a_page="""\
#########
A section
#########

A plot that does not show its source (but does get run by doctest).

.. nbplot::
    :include-source: false

    >>> a = 1

The next incantation is nearly the same, except we also show the output in the
text builder (but no other).

.. nbplot::
    :hide-from: all
    :show-to: doctest text

    >>> b = 2

Here we hide the output from everything, including doctests, but show it to the
text builder again.

.. nbplot::
    :hide-from: all
    :show-to: text

    # Enigmatic sentence.

Show to everything (including doctest builder).

.. nbplot::

    >>> assert a == 1
    >>> assert b == 2

Hide from doctest builder, but no other.

.. nbplot::
    :hide-from: doctest

    >>> a = 99

Show that the doctest builder did not see the previous plot directive.

.. nbplot::

    >>> a == 1
    True
""")

    def test_hide_show(self):
        built = self.get_built_file('a_page.txt')
        expected = """\
A section
*********

A plot that does not show its source (but does get run by doctest).

The next incantation is nearly the same, except we also show the
output in the text builder (but no other).

>>> b = 2

Here we hide the output from everything, including doctests, but show
it to the text builder again.

   # Enigmatic sentence.

Show to everything (including doctest builder).

>>> assert a == 1
>>> assert b == 2

Hide from doctest builder, but no other.

>>> a = 99

Show that the doctest builder did not see the previous plot directive.

>>> a == 1
True
"""
        # Blank lines at beginning differ in Sphinx versions
        assert built.strip() == expected.strip()


class TestHideShowTests(TestHideShow):
    """ Test that the doctests pass, requining the hide/show to work
    """

    builder = 'doctest'

    def test_hide_show(self):
        # The build tests the doctests - here we test the doctree
        built = self.get_doctree('a_page')
        expected_node_types = [title] + [
            paragraph,
            nbplot_container,
            nbplot_epilogue] * 4
        for node, exp_type in zip(built[0].children, expected_node_types):
            assert type(node) == exp_type


class TestHideShowHtml(TestHideShow):
    """ Test that the HTML does not see the hidden section
    """

    builder = 'html'

    def test_hide_show(self):
        built = self.get_built_file('a_page.html')
        assert not '# Enigmatic sentence' in built



""" Testing from_notebook module
"""

from os.path import dirname, join as pjoin

from ..ipython_shim import nbformat
from ..from_notebook import (convert_nb, convert_nb_fname, to_doctests,
                             has_mpl_inline, CODE_WITH_OUTPUT)


DATA_PATH = pjoin(dirname(__file__), 'data')

PLT_HDR = "\n.. nbplot::\n\n"


def test_simple_cells():
    v4 = nbformat.v4
    nb = v4.new_notebook()
    # Markdown -> default conversion
    md_cell = v4.new_markdown_cell('# Some text')
    nb['cells'] = [md_cell]
    exp_text = "\nSome text\n=========\n"
    assert convert_nb(nb) == exp_text
    # Code -> replaced with plot directive / doctest markers
    code_cell = v4.new_code_cell('a = 10')
    nb['cells'] = [code_cell]
    exp_code = PLT_HDR + "    >>> a = 10\n"
    assert convert_nb(nb) == exp_code
    # Empty code -> no output
    empty_code_cell = v4.new_code_cell('')
    nb['cells'] = [empty_code_cell]
    exp_empty_code = "\n"
    # nbconvert 5.3.0 started returning '' from an empty code cell
    assert convert_nb(nb) in (exp_empty_code, '')
    # magic lines get stripped
    magic_code_cell = v4.new_code_cell('%timeit a = 1')
    nb['cells'] = [magic_code_cell]
    assert convert_nb(nb) == exp_empty_code
    # Magic lines stripped from within other code lines
    mixed_magic_code_cell = v4.new_code_cell('%timeit a = 1\nb = 2')
    exp_mixed_magic = PLT_HDR + "    >>> b = 2\n"
    nb['cells'] = [mixed_magic_code_cell]
    assert convert_nb(nb) == exp_mixed_magic


def test_mpl_inline_works():
    # Test we get a mpl-interactive directive for %matplotlib inline
    v4 = nbformat.v4
    nb = v4.new_notebook()
    code_cell = v4.new_code_cell('%matplotlib inline\na = 10')
    nb['cells'] = [code_cell]
    exp_code = "\n.. mpl-interactive::\n{}    >>> a = 10\n".format(PLT_HDR)
    assert convert_nb(nb) == exp_code


def test_mpl_inline():
    assert not has_mpl_inline('')
    assert not has_mpl_inline('%matplotlib inline # foo')
    assert has_mpl_inline('%matplotlib inline')
    assert has_mpl_inline('%   matplotlib    inline')
    assert has_mpl_inline('  %   matplotlib    inline')
    assert not has_mpl_inline('%matplotlib nbagg # foo')
    assert has_mpl_inline('%matplotlib nbagg')
    assert has_mpl_inline('% matplotlib nbagg\nb = 2')
    assert has_mpl_inline('a = 1\n% matplotlib nbagg\nb = 2')


def test_to_doctests():
    # Test to_doctests filter
    assert to_doctests('') == ''
    assert to_doctests('a = 1') == '>>> a = 1'
    assert to_doctests('a = 1\nb = 2') == '>>> a = 1\n>>> b = 2'
    assert (to_doctests(
"""
a = 1
for i in (1, 2):
    a += i

    for j in (2, 3):
        a += j

print(a)
for i in (1, 2):

    a += i
print(a)
""") ==
""">>>
>>> a = 1
>>> for i in (1, 2):
...     a += i
...
...     for j in (2, 3):
...         a += j
...
>>> print(a)
>>> for i in (1, 2):
...
...     a += i
>>> print(a)
""")
    assert (to_doctests(
"""def xyz_trans_vol(vol, x_y_z_trans):
    \"\"\" Make a new copy of `vol` translated by `x_y_z_trans` voxels

    x_y_z_trans is a sequence or array length 3, containing the (x, y, z) translations in voxels.

    Values in `x_y_z_trans` can be positive or negative, and can be floats.
    \"\"\"
    x_y_z_trans = np.array(x_y_z_trans)
""") ==
""">>> def xyz_trans_vol(vol, x_y_z_trans):
...     \"\"\" Make a new copy of `vol` translated by `x_y_z_trans` voxels
...
...     x_y_z_trans is a sequence or array length 3, containing the (x, y, z) translations in voxels.
...
...     Values in `x_y_z_trans` can be positive or negative, and can be floats.
...     \"\"\"
...     x_y_z_trans = np.array(x_y_z_trans)
""")


def test_small():
    # Regression tests on small notebook
    nb_fname = pjoin(DATA_PATH, 'small.ipynb')
    rst_fname = pjoin(DATA_PATH, 'small.rst')
    out = convert_nb_fname(nb_fname)
    with open(rst_fname, 'rt') as fobj:
        assert out + '\n' == fobj.read()


code = \
"""##CODE_START##
a = 1
b = 2
##CODE_END##"""

code_value = 'a = 1\nb = 2\n'

stdout = \
"""##STDOUT_START##
one
two
##STDOUT_END##"""

stdout_value = 'one\ntwo\n'

end_out = \
"""##END_OUT_START##
three
4
##END_OUT_END##"""

end_out_value = 'three\n4\n'


def test_code_regex():
    # Test regular expression matching for code, output parts

    def get_dict(s):
        match = CODE_WITH_OUTPUT.search(s)
        if match is None:
            return None
        return match.groupdict()

    # Check that the code part must match
    assert get_dict(stdout) is None
    assert get_dict(end_out) is None
    # Check different joins still allows detection of parts
    for combination, output in zip(
        ((code,),
         (code, stdout),
         (code, end_out),
         (code, stdout, end_out)),
        (dict(code=code_value, stdout=None, end_out=None),
         dict(code=code_value, stdout=stdout_value, end_out=None),
         dict(code=code_value, stdout=None, end_out=end_out_value),
         dict(code=code_value, stdout=stdout_value, end_out=end_out_value)
        )):
        for joiner in ('\n', '\n  \n\n'):
            in_str = joiner.join(combination)
            assert get_dict(in_str) == output
            # Check adding extra carriage returns etc is OK
            in_str += '\n\n'
            assert get_dict(in_str) == output



""" Test bare Converter class
"""

import re

from nb2plots.converters import Converter


NEW_PAGE = u"""
More fancy title
++++++++++++++++

More compelling text
"""

def test_converter():
    # Default converter
    conv = Converter()
    text = conv.from_rst(NEW_PAGE)
    assert text.strip() == """\
More fancy title
****************

More compelling text"""
    # pseudoxml converter
    conv = Converter('pseudoxml')
    pxml = conv.from_rst(NEW_PAGE)
    assert re.search(r"""<document source=".*/contents.rst">
    <section ids="more-fancy-title" names="more\\ fancy\\ title">
        <title>
            More fancy title
        <paragraph>
            More compelling text
""", pxml) is not None



# ## Another title
#
# Here a `literal` followed by a:
#
# ```
# Literal block, not :math:`parsed`.
# ```
#
# On it goes, to another:
#
# ```
# This block has a very long line, that goes beyond the normal 80-or-so characters, to see whether not-wrapping works correctly.
# ```
#
# Finishes with dramatic exhortation.



""" Convert ReST and doctrees to various formats using nb2plots defaults
"""

import sys
from os.path import join as pjoin
from importlib import import_module

from docutils.io import Output

from sphinxtesters import TempApp


class UnicodeOutput(Output):
    """ Don't do anything to the string; just return it.
    """

    default_destination_path = '<string>'

    def write(self, data):
        """ Store `data` in `self.destination`, and return it."""
        self.destination = data
        return data


class Converter(object):
    """ Class to convert from ReST or doctree to output format
    """

    default_conf = ''

    def __init__(self, buildername='text', conf_txt=None,
                 status=sys.stdout, warningiserror=True):
        """ Build ReST text in string `rst_text` into doctree.

        Parameters
        ----------
        buildername : str, optional
            Builder name.
        conf_txt : None or str
            Text for ``conf.py`` file controlling Sphinx build.
        status : file-like object or None, optional
            File-like object to which to write build status messages, or None
            for no build status messages.
        warningiserror : {True, False}, optional
            if True, raise an error for warning during the Sphinx build.
        """
        self.buildername = buildername
        self.conf_txt = conf_txt if not conf_txt is None else self.default_conf
        self.status = status
        self.warningiserror = warningiserror

    def _make_app(self, rst_text):
        """ Make, return Sphinx application instance for input ReST text.
        """
        return TempApp(rst_text,
                       self.conf_txt, status=self.status,
                       warningiserror=self.warningiserror,
                       buildername=self.buildername)

    def _build_rst(self, rst_text, resolve=True):
        """ Build ReST text in string `rst_text` into doctree.

        Parameters
        ----------
        rst_text : str
            string containing ReST to build.
        resolve : {True, False}, optional
            Whether to resolve references before returning doctree.

        Returns
        -------
        doctree : node
            document node.
        app : object
            Sphinx application object.  This will need to be cleaned up
            (``app.cleanup()``) after use.
        """
        app = self._make_app(rst_text)
        out_fname = pjoin(app.tmp_dir, 'contents.rst')
        with open(out_fname, 'wt') as fobj:
            fobj.write(rst_text)
        # Force build of everything
        app.build(True, [])
        if resolve:
            dt = app.env.get_and_resolve_doctree('contents', app.builder)
        else:
            dt = app.env.get_doctree('contents')
        return dt, app

    def from_doctree(self, doctree, builder):
        """ Convert doctree `doctree` to output format

        Parameters
        ----------
        doctree : node
            Document node.
        builder : object
            Sphinx builder object.

        Returns
        ------
        output : str
            Representation in output format
        """
        builder.prepare_writing(['contents'])
        return builder.writer.write(doctree, UnicodeOutput())

    def from_rst(self, rst_text, resolve=True):
        """ Build Sphinx formatted ReST text `rst_text` into output format

        Parameters
        ----------
        rst_text : str
            string containing ReST to build.
        resolve : {True, False}, optional
            Whether to resolve references before returning doctree.

        Returns
        -------
        output : str
            Text in output format
        """
        doctree, app = self._build_rst(rst_text, resolve)
        res = self.from_doctree(doctree, app.builder)
        app.cleanup()
        return res


def can_import(module_str):
    try:
        import_module(module_str)
    except ImportError:
        return False
    return True


DEFAULT_EXTENSIONS = [ext_name for ext_name in
                      ["nb2plots",
                       'sphinx.ext.autodoc',  # to silence math_dollar warning
                       'sphinx.ext.mathjax',  # to enable math output
                       'texext.math_dollar']  # to enable inline dollar syntax
                      if can_import(ext_name)]


DEFAULT_CONF =  """\
extensions = [{}]
""".format(',\n'.join('"{}"'.format(ext_name)
                      for ext_name in DEFAULT_EXTENSIONS))


class NbConverter(Converter):
    default_conf = DEFAULT_CONF


# Some standard converters
to_pxml = NbConverter('pseudoxml')
to_markdown = NbConverter('markdown')
to_py = NbConverter('python')
to_notebook = NbConverter('jupyter')



# 1. First ordered list item
#
# 1. Another item
#
#    * Unordered sub-list.
#
#    * Another sub-item.
#
# 1. Actual numbers don't matter, just that it's a number
#
#    1. Ordered sub-list
#
# 1. And another item.
#
#    You can have properly indented paragraphs within list items. Notice the
#    blank line above.  The text can keep going on for a long time, flowing
#    across
#    *many* lines. There may be stuff such as $a = b$ and `literals`.
#
#    You can have a line break within the list.
#
#    Note that this line is separate, but within the same paragraph.



# ## Nbplot file
#
# Introducing...

a = 1
a

# ... the `nbplot`.

b = 1
a

# If running in the IPython console, consider running `%matplotlib` to enable
# interactive plots.  If running in the Jupyter Notebook, use `%matplotlib
# inline`.
#
# A block to test line-wrapping:

# Here is a comment line that goes beyond the normal 80 character line wrap to test (non) wrapping

# Markdown builder can be forced to skip:

# Should not appear in Markdown

# Python builder can be forced to skip:
#
# Jupyter builder can be forced to skip:

# Should not appear in Notebook



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Test scripts

Test running scripts
"""
from __future__ import division, print_function, absolute_import

from os.path import (join as pjoin, exists)
from glob import glob
import re

from scripttester import ScriptTester

from nb2plots.testing import DATA_PATH
from nb2plots.testing.convutils import fcontents, unsmart, unsmart_nb
from nb2plots.testing.nbtesters import assert_nb_equiv


runner = ScriptTester('nb2plots', win_bin_ext='.bat')
run_command = runner.run_command


def script_test(func):
    # Decorator to label test as a script_test
    func.script_test = True
    return func


@script_test
def test_rst2md():
    # test rst2md script over all .rst files checking against .md files
    for rst_fname in glob(pjoin(DATA_PATH, '*.rst')):
        md_fname = rst_fname[:-3] + 'md'
        with open(md_fname, 'rb') as fobj:
            expected_md = fobj.read()
        # Skip files containing text "skip".  These are files for which the
        # source ReST is not valid in plain docutils, such as those containing
        # Sphinx directives and roles.
        if expected_md.strip() == b'skip':
            continue
        cmd = ['rst2md', rst_fname]
        code, stdout, stderr = run_command(cmd)
        assert stdout == expected_md


@script_test
def test_sphinx2md():
    # test sphinx2md script over all .rst files checking against .smd / .md
    # files
    for rst_fname in glob(pjoin(DATA_PATH, '*.rst')):
        # Try .smd filename first, otherwise ordinary .md
        md_fname = rst_fname[:-3] + 'smd'
        if not exists(md_fname):
            md_fname = rst_fname[:-3] + 'md'
        expected_md = fcontents(md_fname)
        cmd = ['sphinx2md', rst_fname]
        code, stdout, stderr = run_command(cmd)
        assert (unsmart(stdout.decode('utf-8')) ==
                expected_md.decode('utf-8'))


@script_test
def test_sphinx2nb():
    # test sphinx2nb script over all .rst files checking against .ipynb files
    for rst_fname in glob(pjoin(DATA_PATH, '*.rst')):
        nb_fname = rst_fname[:-3] + 'ipynb'
        expected = fcontents(nb_fname, 't')
        cmd = ['sphinx2nb', rst_fname]
        code, stdout, stderr = run_command(cmd)
        assert_nb_equiv(unsmart_nb(stdout.decode('utf-8')),
                        expected)


@script_test
def test_sphinx2py():
    # test sphinx2py script over all .rst files checking against .ipynb files
    for rst_fname in glob(pjoin(DATA_PATH, '*.rst')):
        py_fname = rst_fname[:-3] + 'py'
        expected = fcontents(py_fname, 'b')
        cmd = ['sphinx2py', rst_fname]
        code, stdout, stderr = run_command(cmd)
        assert (unsmart(stdout.decode('utf-8')) ==
                expected.decode('utf-8'))


@script_test
def test_sphinx2pxml():
    rst_fname = pjoin(DATA_PATH, 'sect_text.rst')
    cmd = ['sphinx2pxml', rst_fname]
    code, stdout, stderr = run_command(cmd)
    pattern = r"""<document source=".*?">
    <section ids="a-section" names="a\\ section">
        <title>
            A section
        <paragraph>
            Some 
            <emphasis>
                text
            ."""
    output = stdout.decode('utf-8')
    assert re.match(pattern, output)



""" Directive to handle ``%matplotlib [inline]`` in Sphinx, Notebooks

The directive is a marker to tell the Notebook converter to put a ``%matplotlib
inline`` code cell at this position in the text.  Use thusly::

    .. mpl-interactive::

With no content (as above), the directive inserts a Hint into the ReST text
suggesting ``%matplotlib`` in the IPython console, or ``%matplotlib inline`` in
the Notebook.  You can specify some other message with::

    .. mpl-interactive::

        Your text here.

The text can be any valid ReStructured text.
"""

from docutils.parsers.rst.directives.admonitions import Hint
from docutils import nodes
from docutils.statemachine import StringList


class mpl_hint(nodes.hint):
    """ Node signals presence of MPL interactive block """


class MPLInteractive(Hint):
    """ Output information on IPython / MPL interactive mode in Sphinx
    """

    node_class = mpl_hint

    default_text = """\
If running in the IPython console, consider running ``%matplotlib`` to enable
interactive plots.  If running in the Jupyter Notebook, use ``%matplotlib
inline``.
"""

    def run(self):
        if len(self.content) == 0:
            self.content = StringList(self.default_text.splitlines())
        return super(MPLInteractive, self).run()


def visit_mpl_inter(self, node):
    self.visit_hint(node)


def depart_mpl_inter(self, node):
    self.depart_hint(node)


def setup(app):
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    # mpl_hint treated same as hint
    app.add_node(mpl_hint,
                 **{builder: (visit_mpl_inter, depart_mpl_inter)
                    for builder in ('html', 'latex', 'text')})
    app.add_directive('mpl-interactive', MPLInteractive)



""" Sphinx builders for output formats
"""

from sphinx.builders.text import TextBuilder

from . import doctree2md, doctree2py, doctree2nb


class MarkdownBuilder(TextBuilder):
    name = 'markdown'
    format = 'markdown'
    out_suffix = '.md'
    writer_class = doctree2md.Writer

    def __init__(self, app):
        """ Initialize Markdown (and friends) builder
        """
        super(MarkdownBuilder, self).__init__(app)
        # If None or empty string, do not resolve internal links.  Otherwise,
        # use as base for HTML-style links.  Applies to internal references and
        # download references.
        self.markdown_http_base = (self.config.markdown_http_base
                                   if self.config.markdown_http_base
                                   else None)

    def prepare_writing(self, docnames):
        self.writer = self.writer_class(self)

    def get_target_uri(self, docname, typ=None):
        if self.markdown_http_base:
            return docname + '.html'
        return super(MarkdownBuilder, self).get_target_uri(docname, typ)


class PythonBuilder(MarkdownBuilder):
    name = 'python'
    out_suffix = '.py'
    writer_class = doctree2py.Writer


class NotebookBuilder(PythonBuilder):
    name = 'jupyter'
    out_suffix = '.ipynb'
    writer_class = doctree2nb.Writer


def setup(app):
    app.add_builder(MarkdownBuilder)
    app.add_builder(PythonBuilder)
    app.add_builder(NotebookBuilder)
    # Base URL for Markdown link conversion
    app.add_config_value('markdown_http_base', None, True)



# > Blockquotes are very handy in email to emulate reply text.
# > This line is part of the same quote.
#
# Quote break.
#
# > This is a very long line that will still be quoted properly when it wraps.
# > Oh boy let's keep writing to make sure this is long enough to actually wrap
# > for everyone. Oh, you can *put* **Markdown** into a blockquote.



""" Test conversion of doctree to Jupyter notebook
"""
from os.path import join as pjoin
from glob import glob

from nb2plots.converters import to_notebook
from nb2plots.ipython_shim import nbf

# Shortcuts
n_nb = nbf.new_notebook
n_md_c = nbf.new_markdown_cell
n_c_c = nbf.new_code_cell

from nb2plots.testing import DATA_PATH
from nb2plots.testing.convutils import fcontents, unsmart_nb
from nb2plots.testing.nbtesters import assert_nb_equiv


def to_nb_safe(rst_str):
    out = to_notebook.from_rst(rst_str)
    return unsmart_nb(out)


def cells2json(cells):
    nb = nbf.new_notebook()
    nb['cells'] += cells
    return nbf.writes(nb)


def assert_rst_cells_equal(rst_text, cells):
    actual = to_notebook.from_rst(rst_text)
    expected = cells2json(cells)
    assert actual == expected


def test_basic():
    assert_rst_cells_equal('Some text', [n_md_c('Some text')])


def test_runrole_reference():
    # Ignore notebook reference in source ReST
    assert_rst_cells_equal('Some text :clearnotebook:`.`',
                           [n_md_c('Some text')])
    assert_rst_cells_equal('Some text :fullnotebook:`.`',
                           [n_md_c('Some text')])


def test_only():
    for builder_name in ('html', 'latex', 'unbelievable'):
        assert_rst_cells_equal(
"""
Before

.. only:: {0}

    Specific to builder {0}

After""".format(builder_name),
            [n_md_c('Before\n\nAfter')])
    assert_rst_cells_equal(
"""
Before

.. only:: markdown

    More text

After""".format(builder_name),
            [n_md_c('Before\n\nMore text\n\nAfter')])


def test_doctests():
    assert_rst_cells_equal("""\
Text 1

>>> # A comment
>>> a = 1

Text 2
""", [n_md_c('Text 1'), n_c_c('# A comment\na = 1'), n_md_c('Text 2')])


def test_nbplots():
    # nbplot directive with doctest markers
    assert_rst_cells_equal("""\
Text 1

.. nbplot::

    >>> # A comment
    >>> a = 1

Text 2
""", [n_md_c('Text 1'), n_c_c('# A comment\na = 1'), n_md_c('Text 2')])
    # nbplot directive with no doctest markers
    assert_rst_cells_equal("""\
Text 1

.. nbplot::

    # A comment
    a = 1

Text 2
""", [n_md_c('Text 1'), n_c_c('# A comment\na = 1'), n_md_c('Text 2')])
    # Doctest interspersed with text
    assert_rst_cells_equal("""\
Text 1

.. nbplot::

    >>> # A comment

    Some thoughts I had

    >>> a = 1

Text 2
""", [n_md_c('Text 1'),
      n_c_c('# A comment'),
      n_md_c('Some thoughts I had'),
      n_c_c('a = 1'),
      n_md_c('Text 2')])


def assert_conv_equal(rst_str, expected):
    assert_nb_equiv(to_nb_safe(rst_str), expected)


def test_example_files():
    # test conversion over all .rst files, checking against .ipynb files
    for rst_fname in glob(pjoin(DATA_PATH, '*.rst')):
        rst_contents = fcontents(rst_fname, 't')
        nb_fname = rst_fname[:-3] + 'ipynb'
        nb_contents = fcontents(nb_fname, 't')
        assert_conv_equal(rst_contents, nb_contents)


def test_notebook_basic():
    # Test conversion of basic ReST to ipynb JSON
    ipynb = to_notebook.from_rst(r"""
Title
=====

Some text with :math:`a = 1` math.

.. math::

    \textrm{math block}

.. nbplot::

    >>> c = 1
    >>> c
    1

More text.

.. nbplot::

    >>> d = 2
    >>> d
    2""")
    expected = r"""{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Title\n",
    "\n",
    "Some text with $a = 1$ math.\n",
    "\n",
    "$$\n",
    "\\textrm{math block}\n",
    "$$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "c = 1\n",
    "c"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "More text."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "d = 2\n",
    "d"
   ]
  }
 ],
 "metadata": {},
 "nbformat": 4,
 "nbformat_minor": 1
}"""
    assert_nb_equiv(ipynb, expected)


def test_default_mathdollar():
    # Test mathdollar extension present by default.
    ipynb = to_notebook.from_rst(r'Some text with $a = 1$ math.')
    expected = r"""{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Some text with $a = 1$ math."
   ]
  }
 ],
 "metadata": {},
 "nbformat": 4,
 "nbformat_minor": 1
}"""
    assert_nb_equiv(ipynb, expected)





""" Test conversion of Sphinx format ReST to Markdown

Test running writer over example files and chosen snippets.
"""
from __future__ import (division, print_function, absolute_import,
                        unicode_literals)

from os.path import join as pjoin, exists
from glob import glob

from nb2plots.converters import to_markdown

from nb2plots.testing import DATA_PATH
from nb2plots.testing.convutils import (convert_assert, fcontents,
                                        unsmart_converter)


to_md_safe = unsmart_converter(to_markdown.from_rst)


def assert_conv_equal(rst_str, md_expected):
    convert_assert(rst_str, to_md_safe, md_expected, None)


def test_example_files():
    # test sphinx2md script over all .rst files checking against .smd / .md
    # files
    for rst_fname in glob(pjoin(DATA_PATH, '*.rst')):
        rst_contents = fcontents(rst_fname, 't')
        # Try .smd filename first, otherwise ordinary .md
        md_fname = rst_fname[:-3] + 'smd'
        if not exists(md_fname):
            md_fname = rst_fname[:-3] + 'md'
        md_contents = fcontents(md_fname, 't')
        assert_conv_equal(rst_contents, md_contents)


def test_default_mathdollar():
    # Test mathdollar extension present by default.
    md = to_markdown.from_rst(r'Some text with $a = 1$ math.')
    expected = "Some text with $a = 1$ math.\n"
    assert md == expected



# -*- coding: utf-8 -*-
""" Tests for nb2plots.nbplot extension
"""

from nb2plots import nbplots as nbp

from nb2plots.testing import mockapp


def test_nbplots_setup():
    # Test extension setup works as expected
    app = mockapp.get_app()
    nbp.setup(app)
    config_names = ['nbplot_pre_code',
                    'nbplot_include_source',
                    'nbplot_formats',
                    'nbplot_html_show_formats',
                    'nbplot_rcparams',
                    'nbplot_working_directory',
                    'nbplot_template',
                    'nbplot_flags']
    connects = [
        ('builder-inited', nbp.do_builder_init),
        ('env-purge-doc', nbp.do_purge_doc),
    ]
    for method_name, args, kwargs in app.method_calls:
        if (method_name == 'add_config_value' and
            args[0] in config_names):
            config_names.remove(args[0])
        if (method_name == 'connect' and args[0:2] in connects):
            connects.remove(args[0:2])
    assert len(config_names) == 0, \
            'config set failed for {}'.format(config_names)
    assert len(connects) == 0, 'Connections failed'



""" Tests for proj1 build using sphinx extensions """

from os.path import (join as pjoin, dirname, isdir, exists)

from sphinxtesters import ModifiedPageBuilder

HERE = dirname(__file__)

PAGE_HEADER = """\
A title
=======

"""

class Proj1Builder(ModifiedPageBuilder):
    """ Build using 'proj1' directory as template to modify
    """

    page_source_template = pjoin(HERE, 'proj1')

    # default_page used in 'replace_page' class method
    default_page = 'a_page.rst'


class TestProj1(Proj1Builder):

    def test_basic_build(self):
        assert isdir(self.out_dir)
        assert isdir(self.doctree_dir)
        doctree = self.get_doctree('a_page')
        assert len(doctree.document) == 1
        doctree_str = self.doctree2str(doctree)
        expected = (
            '<title>A section</title>\n'
            '<paragraph>Some text.</paragraph>\n'
            '<paragraph><runrole_reference '
            'refdoc="a_page" reftarget="/a_page.ipynb" reftype="clearnotebook">'
            'notebook here</runrole_reference></paragraph>\n'
            '<paragraph>More text.</paragraph>\n'
            '<paragraph><runrole_reference '
            'refdoc="a_page" reftarget="another.ipynb" reftype="fullnotebook">'
            'full</runrole_reference></paragraph>\n'
            '<paragraph>Text is endless.</paragraph>\n'
            '<paragraph><runrole_reference '
            'refdoc="a_page" reftarget="/a_page.py" reftype="pyfile">'
            'code here</runrole_reference></paragraph>')
        assert doctree_str == expected
        # Check the expected files were written
        for fname in ('a_page.ipynb', 'another.ipynb', 'a_page.py'):
            built_fname = pjoin(self.build_path, 'html', fname)
            assert exists(built_fname)


class TestNotSameName(Proj1Builder):

    @classmethod
    def modify_source(cls):
        with open(pjoin(cls.page_source, 'a_page.rst'), 'wt') as fobj:
            fobj.write(PAGE_HEADER +
"""
:clearnotebook:`.` (the default name and text).

:fullnotebook:`Get it here <another_name.ipynb>` (not-default name).
""")


class TestSameNameIpy(Proj1Builder):
    should_error = True

    @classmethod
    def modify_source(cls):
        with open(pjoin(cls.page_source, 'a_page.rst'), 'wt') as fobj:
            fobj.write(PAGE_HEADER +
"""
:clearnotebook:`.` (the default name).

:fullnotebook:`.` (default name again).
""")

class TestSameNamePy(Proj1Builder):
    should_error = True

    @classmethod
    def modify_source(cls):
        with open(pjoin(cls.page_source, 'a_page.rst'), 'wt') as fobj:
            fobj.write(PAGE_HEADER +
"""
:clearnotebook:`.` (the default name).

:pyfile:`code file <a_page.ipynb>` (same name as notebook)
""")



#!python
""" Convert from notebook to rst page with plot directives
"""

import re

from jinja2 import DictLoader

from .ipython_shim import nbformat, nbconvert, traitlets, config

MPL_INLINE = re.compile(r"^\s*%\s*matplotlib\s+(inline|nbagg)\s*$",
                        re.MULTILINE)


# Template to label code and output and plot blocks
dl = DictLoader({'rst_plots.tpl': """\
{%- extends 'rst.tpl' -%}

{% block input %}
{%- if cell.source.strip() | has_mpl_inline -%}
.. mpl-interactive::

{% endif -%}
{%- if cell.source.strip() | strip_ipy -%}
##CODE_START##
{{ cell.source | strip_ipy | to_doctests | indent}}
##CODE_END##
{% endif -%}
{% endblock input %}

{%- block execute_result -%}
{%- block data_priority scoped -%}
{{ super() }}
{%- endblock -%}
{%- endblock execute_result -%}

{%- block data_svg -%}
{%- endblock data_svg -%}
{%- block data_png -%}
{%- endblock data_png -%}

{%- block stream -%}
##STDOUT_START##
{{ output.text | ellipse_mpl | indent }}
##STDOUT_END##
{%- endblock stream -%}

{%- block data_text scoped -%}
##END_OUT_START##
{{ output.data['text/plain'] | ellipse_mpl | indent }}
##END_OUT_END##
{%- endblock data_text -%}
"""})


def has_mpl_inline(code):
    return MPL_INLINE.search(code)


def strip_ipy(code):
    """ Strip any ipython magic lines """
    return '\n'.join([line for line in code.split('\n')
                      if not line.strip().startswith('%')])


def to_doctests(code, first='>>> ', cont='... '):
    """Add docstring prompts to code snippets"""
    new_code = []
    code_list = code.split('\n')
    prefix = first
    last_line_no = len(code_list) - 1
    for i, line in enumerate(code_list):
        if line.strip() != '':
            prefix = cont if line.startswith(' ') else first
            new_code.append(prefix + line)
            continue
        # For blank lines, we always strip the whitespace, but we will need a
        # prefix for anything but the last line
        if i == last_line_no:
            prefix = ''
        elif code_list[i + 1].startswith(' '):
            # Use continuation for line following
            prefix = cont.rstrip()
        else:  # Use prefix for previous line
            prefix = prefix.rstrip()
        new_code.append(prefix)

    return '\n'.join(new_code)


MPL_LIST_OUT = re.compile('\[<matplotlib\..*?>\]')
MPL_OBJ_OUT = re.compile('<matplotlib\..*?>')

def ellipse_mpl(text):
    """ Replace outputs of matplotlib objects with ellipses
    """
    text = MPL_LIST_OUT.sub('[...]', text)
    return MPL_OBJ_OUT.sub('<...>', text)


class PlotsExporter(nbconvert.RSTExporter):
    template_file = 'rst_plots.tpl'
    filters = traitlets.Dict(dict(
        has_mpl_inline=has_mpl_inline,
        to_doctests=to_doctests,
        strip_ipy=strip_ipy,
        ellipse_mpl=ellipse_mpl,
    ), config=True)


# Code, with option stdout and optional end-of-block output
CODE_WITH_OUTPUT = re.compile(
    '^##CODE_START##\n'
    '(?P<code>.*?)'
    '^##CODE_END##(\n|$)'
    '([\s\\n]*?'
    '^##STDOUT_START##\n'
    '(?P<stdout>.*?)'
    '^##STDOUT_END##(\n|$))?'
    '([\s\\n]*?'
    '^##END_OUT_START##\n'
    '(?P<end_out>.*?)'
    '^##END_OUT_END##(\n|$))?', re.S | re.M)


PLOT_DIRECTIVE_PREFIX = """\
.. nbplot::

"""

def repl_code_plot(match):
    groups = match.groupdict(default='')
    out = ''.join((PLOT_DIRECTIVE_PREFIX,
                   groups['code'],
                   groups['end_out']))
    if groups['stdout']:
        out += '\n' + groups['stdout']
    return out


def convert_nb_fname(nb_fname):
    with open(nb_fname, 'rt') as nb_fobj:
        notebook = nbformat.read(nb_fobj, as_version=4)
    return convert_nb(notebook)


def convert_nb(notebook):
    # Turn off output preprocessor (we don't want the figures)
    c =  config.Config({
        'ExtractOutputPreprocessor':{'enabled': False}
    })
    plots_exporter = PlotsExporter(extra_loaders=[dl], config=c)
    output, resources = plots_exporter.from_notebook_node(notebook)
    return CODE_WITH_OUTPUT.sub(repl_code_plot, output)



# nb2plots package

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from . import nbplots
from . import runroles
from . import mpl_interactive
from . import codelinks
from . import sphinx2foos


def setup(app):
    nbplots.setup(app)
    runroles.setup(app)
    mpl_interactive.setup(app)
    codelinks.setup(app)
    sphinx2foos.setup(app)



""" Test conversion of Sphinx format ReST to Python .py files

Test running writer over example files and chosen snippets.
"""
from __future__ import (division, print_function, absolute_import,
                        unicode_literals)

from os.path import join as pjoin
from glob import glob

from nb2plots.doctree2py import parse_doctest

from nb2plots.converters import to_py

from nb2plots.testing import DATA_PATH
from nb2plots.testing.convutils import (convert_assert, fcontents,
                                        unsmart_converter)


def test_doctest_parser():
    assert parse_doctest('>>> # comment') == '# comment'
    assert parse_doctest('>>> a = 10') == 'a = 10'
    assert parse_doctest('   >>> a = 10') == 'a = 10'
    assert (parse_doctest('   >>> a = 10\n   >>> b = 20') ==
                 'a = 10\nb = 20')
    assert (parse_doctest(
        '   >>> for i in (1, 2):\n   ...     print(i)') ==
        'for i in (1, 2):\n    print(i)')


to_py_safe = unsmart_converter(to_py.from_rst)


def assert_conv_equal(rst_str, md_expected, unsmart=True):
    convert_assert(rst_str, to_py_safe, md_expected, None)


def test_example_files():
    # test conversion code over all .rst files checking against .py files
    for rst_fname in glob(pjoin(DATA_PATH, '*.rst')):
        rst_contents = fcontents(rst_fname, 't')
        py_fname = rst_fname[:-3] + 'py'
        contents = fcontents(py_fname, 't')
        assert_conv_equal(rst_contents, contents)



# Here is some *emphasis* and here is an actual \* asterisk.
#
# You can have actual \`backticks\` too.
#
# And real \\backslashes.
#
# No escaping in $1 * 2$.
#
# No escaping in a math block:
#
# $$
# 3 * 4
# $$



"""
A directive for including a matplotlib plot in a Sphinx document.

This directive is based on the ``plot`` directive of matplotlib, with thanks.

By default, in HTML output, `nbplot` will include a .png file with a
link to a high-res .png and .pdf.  In LaTeX output, it will include a
.pdf.

The source code for the plot is **inline content** to the directive:.  It can
be with or without *doctest* syntax.  Here's an example without doctest
syntax::

    .. nbplot::

        import matplotlib.pyplot as plt
        import matplotlib.image as mpimg
        import numpy as np
        img = mpimg.imread('_static/stinkbug.png')
        imgplot = plt.imshow(img)

Here's an example with doctest syntax::

    .. nbplot::

        A plotting example:
        >>> import matplotlib.pyplot as plt
        >>> plt.plot([1,2,3], [4,5,6])  #doctest: +ELLIPSIS
        [...]

Options
-------

The ``nbplot`` directive supports the following options:

    format : {'python', 'doctest'}
        Specify the format of the input.  If not specified, nbplot guesses the
        format from the content.

    include-source : bool
        Whether to display the source code. The default can be changed using
        the `nbplot_include_source` variable in ``conf.py``.  Any doctests in
        the source code will still be run by the doctest builder.
        `include-source` has the same effect as ``:hide-from: all`` then
        ``:show-to: doctest`` (in fact, that is how it is implemented).  See
        below for interactions of ``:include-source:`` and these options.

    hide-from : str
        Space-separated list of builders that should not render any source from
        this directive.  Can also be the special value "all", hiding the source
        from all builders.  Any builder names in `show-to` override builders
        specified with this option.  If you apply ``:hide-from:`` options as
        well as ``:include-source: false`` (see above), these will have no
        effect, because ``:include-source: false`` implies ``:hide-from: all``.

    show-to : str
        Space-separated list of builders that *should* render any source from
        this directive.  Builder names here override any builders specified in
        `hide-from`.  If you apply ``:show-to:`` options as well as
        ``:include-source: false`` (see above) these can enable additional
        builders on top of the implied ``doctest`` builder.

    encoding : str
        If this source file is in a non-UTF8 or non-ASCII encoding,
        the encoding must be specified using the `:encoding:` option.
        The encoding will not be inferred using the ``-*- coding -*-``
        metacomment.

    keepfigs : bool
        If specified, do not close figures from previous plot command. This
        allows you to build up a plot in stages, with the current nbplot
        command building on the figure generated by previous nbplot commands.

    nofigs : bool
        If specified, the code block will be run, but no figures will be
        inserted.  This is sometimes necessary when your code generates a plot,
        but you do not want to include it.  If the code does not generate a
        plot, this option is not necessary.

    raises : str
        String giving error class.  The code runner will assert this error was
        raised by the enclosed code, and suppress stderr.

    render-parts : str
        The contents of the directive may be divided into "parts" at lines
        containing ``.. part`` with a blank line before and after.  This option
        gives a Python expression returning an integer or tuple giving the
        indices of the *parts* that should be used for the built output of the
        code block.  These to-be-built parts can be different from the parts
        used to generate the figures associated with the block, defined in the
        ``run-parts`` option below.  The typical use is for building when
        running the ``render-parts`` would be too slow, or the building system
        does not have the requirements to run the code in ``render-parts``.
        The contents of this option is a Python expression returning an integer
        or a tuple, where an integer is the index of the part to render, and a
        tuple is a tuple of such integers. The expression operates in a
        namespace defined by the contents of the ``nbplot_flags`` config
        variable (a dict) and any extra namespace defined above this directive
        with :class:`NBPlotFlags` directives.

        Any selected parts get labeled such that the doctest builder does not
        see them, and therefore they do not get picked up by the sphinx doctest
        extension.  The nbplots extension tells the other builders to build the
        skipped doctest as if it were a standard doctest.

        Examples::

            :render-parts: 0 if have_matlab else 1

        Default value is 0.

    run-parts : str
        See ``render-parts`` above.  Python expression that returns integer or
        tuple giving indices for parts that should be executed when generating
        the figures.  Any doctests in these parts also get wrapped in standard
        doctest blocks, and so will be picked up by the sphinx doctest builder.

        Examples::

            :run-parts: 0 if slow else 1

        Default value is 0.

The namespace of the nbplot command is reset to empty for each document.  The
code in each nbplot directive instance in a given document uses the namespace
generated by previous nbplot directive instances in the same document.

Additionally, this directive supports all of the options of the
`image` directive, except for `target` (since plot will add its own
target).  These include `alt`, `height`, `width`, `scale`, `align` and
`class`.

Configuration options
---------------------

The nbplot directive has the following configuration options:

    nbplot_include_source
        Default value for the include-source option

    nbplot_pre_code
        Code that should be executed before each plot.

    nbplot_formats
        File formats to generate. List of tuples or strings::

            [(suffix, dpi), suffix, ...]

        that determine the file format and the DPI. For entries whose
        DPI was omitted, sensible defaults are chosen. When passing from
        the command line through sphinx_build the list should be passed as
        suffix:dpi,suffix:dpi, ....

    nbplot_html_show_formats
        Whether to show links to the files in HTML.

    nbplot_rcparams
        A dictionary containing any non-standard rcParams that should
        be applied at the beginning of each document.

    nbplot_working_directory
        By default, the working directory will be changed to the directory of
        the example, so the code can get at its data files, if any.  Also its
        path will be added to `sys.path` so it can import any helper modules
        sitting beside it.  This configuration option can be used to specify
        a central directory (also added to `sys.path`) where data files and
        helper modules for all code are located.  If the directory is relative,
        directory taken to be relative to root directory of the project (rather
        than source directory).

    nbplot_template
        Provide a customized template for preparing restructured text.
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import six

from collections import defaultdict, Sequence
import sys, os, shutil, io, re, textwrap
from os.path import (relpath, abspath, join as pjoin, dirname, exists,
                     basename, splitext, isdir)
import traceback
from pprint import pformat

from docutils.statemachine import StringList
from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.images import Image
align = Image.align
from docutils.parsers.rst import Directive

import jinja2
def format_template(template, **kw):
    return jinja2.Template(template).render(**kw)

import matplotlib
import matplotlib.cbook as cbook
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib._pylab_helpers import Gcf

__version__ = 2


class NBPlotFlags(Directive):
    """ Set flag namespace for nbplot
    """
    has_content = True
    required_arguments = 0
    optional_arguments = 0

    def run(self):
        document = self.state.document
        env = document.settings.env
        docname = env.docname
        local_ns = env.nbplot_flag_namespaces[docname]
        six.exec_('\n'.join(self.content), None, local_ns)
        return []


class NBPlotShowFlags(Directive):
    """ Show flag namespace for nbplot

    This directive shows the result of :class:`NBPlotFlags` directives, for
    debugging and testing.
    """
    has_content = True
    required_arguments = 0
    optional_arguments = 0

    def run(self):
        document = self.state.document
        env = document.settings.env
        docname = env.docname
        local_ns = env.nbplot_flag_namespaces[docname]
        content = pformat(local_ns)
        return [nodes.literal_block(content, content)]


# Options for NBPlotDirective

def _option_boolean(arg):
    if not arg or not arg.strip():
        # no argument given, assume used as a flag
        return True
    elif arg.strip().lower() in ('no', '0', 'false'):
        return False
    elif arg.strip().lower() in ('yes', '1', 'true'):
        return True
    else:
        raise PlotValueError('"%s" unknown boolean' % arg)


def _option_format(arg):
    return directives.choice(arg, ('python', 'doctest'))


def _option_align(arg):
    return directives.choice(
        arg,
        ("top", "middle", "bottom", "left", "center", "right"))


class nbplot_container(nodes.container, nodes.Targetable):
    """ Container for rendered nbplot contents

    Also serves as a reference target.
    """

    def likes_builder(self, builder_name):
        """ Return True if attributes specify this is an acceptable build

        Parameters
        ----------
        builder_name : str
            Name of builder by which the node is being visited.

        Returns
        -------
        liked : bool
            True if node attributes are compatible with `builder_name`, False
            otherwise.
        """
        hide_from = self.get('hide-from', [])
        show_to = self.get('show-to', [])
        if builder_name in show_to:
            return True
        if 'all' in hide_from:
            return False
        return builder_name not in hide_from


class nbplot_epilogue(nodes.container, nodes.Targetable):
    """ Container for figures etc following nbplot """


class dont_doctest_block(nodes.General, nodes.FixedTextElement):
    """ Container to hide doctest block from doctest builder """


def doctest_filter(node):
    """ True if node is a ``doctest_block`` node
    """
    return isinstance(node, nodes.doctest_block)


class NBPlotDirective(Directive):
    """ Implements nbplot directive

    """ + __doc__
    has_content = True
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {'alt': directives.unchanged,
                   'height': directives.length_or_unitless,
                   'width': directives.length_or_percentage_or_unitless,
                   'scale': directives.nonnegative_int,
                   'align': _option_align,
                   'class': directives.class_option,
                   'hide-from': directives.unchanged,
                   'show-to': directives.unchanged,
                   'include-source': _option_boolean,
                   'format': _option_format,
                   'keepfigs': directives.flag,
                   'nofigs': directives.flag,
                   'encoding': directives.encoding,
                   'raises': directives.unchanged,
                   'render-parts': directives.unchanged,
                   'run-parts': directives.unchanged,
                  }

    # Node classes for rendered, not rendered nbplot contents
    nbplot_node = nbplot_container
    nbplot_epilogue = nbplot_epilogue
    dont_doctest_block = dont_doctest_block

    def _dont_doctest_doctests(self, tree):
        """ Replace ``doctest_block`` nodes with ``dont_doctest_block`` nodes
        """
        for node in tree.traverse(doctest_filter):
            new_node = self.dont_doctest_block(node.rawsource, node.rawsource)
            node.replace_self(new_node)

    def rst2nodes(self, lines, node_class, node_attrs=None):
        """ Build docutils nodes from list of ReST strings `lines`

        Parameters
        ----------
        lines : list
            list of strings containing ReST.
        node_class : :class:`nodes.Node` class
            Container class for output
        node_attrs : None or dict, optional
            Attributes to apply to instance of `node_class`.  These simply get
            applied to the containing node, except where they signal that the
            contents should be hidden from the doctest builder, in which case
            the doctest_block nodes get swapped for similar nodes that don't
            get picked up by the doctest builder.

        Returns
        -------
        nodes : list
            A length 1 list of nodes, where contained node is of type
            `node_class` and node contents are nodes generated from ReST in
            `lines`.
        """
        node_attrs = {} if node_attrs is None else node_attrs
        text = '\n'.join(lines)
        node = node_class(text)
        for key, value in node_attrs.items():
            if value:
                node[key] = value
        self.add_name(node)
        if len(lines) != 0:
            self.state.nested_parse(StringList(lines),
                                    self.content_offset,
                                    node)
        # Implement hiding of doctests from doctest builder
        if not likes_builder(node, 'doctest'):
            self._dont_doctest_doctests(node)
        return [node]

    def _get_parts(self, option_name):
        if option_name not in self.options:
            return (0,)
        env = self.state.document.settings.env
        name_space = env.nbplot_flag_namespaces[env.docname].copy()
        indices = eval(self.options[option_name], name_space)
        return indices if isinstance(indices, Sequence) else (indices,)

    def _select_parts(self):
        parts = parse_parts(self.content)
        to_render = [parts[i]['contents']
                     for i in self._get_parts('render-parts')]
        to_run = [parts[i]['contents']
                  for i in self._get_parts('run-parts')]
        return ('\n'.join(sum(to_render, [])),
                '\n'.join(sum(to_run, [])))

    def _contains_doctest(self, multi_str):
        """ Check  ``format`` option for doctest specifier, else guess
        """
        if 'format' in self.options:
            return False if self.options['format'] == 'python' else True
        return contains_doctest(multi_str)

    def _build_epilogue(self, images, source_rel_dir, build_dir):
        """ Build the epilogue housing the image links
        """
        document = self.state.document
        config = document.settings.env.config
        nofigs = 'nofigs' in self.options

        # how to link to files from the RST file
        rst_file = document.attributes['source']
        rst_dir = dirname(rst_file)
        dest_dir_link = pjoin(relpath(setup.confdir, rst_dir),
                              source_rel_dir).replace(os.path.sep, '/')
        build_dir_link = relpath(build_dir, rst_dir).replace(os.path.sep, '/')

        images_to_show = [] if nofigs else images[:]
        n_to_show = len(images_to_show)

        opts = [':%s: %s' % (key, val)
                for key, val in self.options.items()
                if key in ('alt', 'height', 'width', 'scale', 'align',
                            'class')]

        # These are the headers for blocks in the template that link to the
        # various build plot images - where the logic is different for each
        # builder.
        only_html = ".. only:: html"
        only_latex = ".. only:: latex"
        only_texinfo = ".. only:: texinfo"

        # Build source text for image link epilogue
        epilogue_source = format_template(
            config.nbplot_template or EPILOGUE_TEMPLATE,
            dest_dir=dest_dir_link,
            build_dir=build_dir_link,
            multi_image=n_to_show > 1,
            only_html=only_html,
            only_latex=only_latex,
            only_texinfo=only_texinfo,
            options=opts,
            images=images_to_show,
            html_show_formats=config.nbplot_html_show_formats and n_to_show)

        return self.rst2nodes(epilogue_source.splitlines(),
                              self.nbplot_epilogue)

    def _copy_image_files(self, images, dest_dir):
        # copy image files to builder's output directory, if necessary
        if not exists(dest_dir):
            cbook.mkdirs(dest_dir)

        for img in images:
            for fn in img.filenames():
                destimg = pjoin(dest_dir, basename(fn))
                if fn != destimg:
                    shutil.copyfile(fn, destimg)

    def _proc_builder_opts(self, config):
        """ Process options related to later per-node selection of builders

        hide-from ad show-to options apply on top of include-source option and
        config setting.
        """
        options = self.options
        options.setdefault('include-source', config.nbplot_include_source)
        if options['include-source']:
            node_attrs = {'hide-from': [], 'show-to': []}
        else:  # include-source is False
            node_attrs = {'hide-from': ['all'], 'show-to': ['doctest']}
        for opt_name in node_attrs:
            if opt_name not in options:
                continue
            values = options[opt_name].split()
            node_attrs[opt_name] += [b_name.strip() for b_name in values]
        return node_attrs

    def run(self):
        document = self.state.document
        config = document.settings.env.config
        env = document.settings.env
        docname = env.docname

        # Check and fill options for skipping nodes per builder
        node_attrs = self._proc_builder_opts(config)

        # If this is the first directive in the document, clear context
        if env.nbplot_reset_markers.get(docname, False):
            context_reset = False
        else:
            env.nbplot_reset_markers[docname] = True
            context_reset = True
        close_figs = False if 'keepfigs' in self.options else True

        counter = document.attributes.get('_plot_counter', 0) + 1
        document.attributes['_plot_counter'] = counter

        source_file_name = document.attributes['source']
        base, ext = splitext(basename(source_file_name))
        output_base = '%s-%d' % (base, counter)
        # ensure that LaTeX includegraphics doesn't choke in foo.bar.pdf
        # filenames
        output_base = output_base.replace('.', '-')

        # should the code raise an exception?
        raises = (eval(self.options['raises']) if 'raises' in self.options
                  else None)

        # determine output directory name fragment
        source_rel_name = relpath(source_file_name, setup.confdir)
        source_rel_dir = dirname(source_rel_name)
        while source_rel_dir.startswith(os.path.sep):
            source_rel_dir = source_rel_dir[1:]

        # build_dir: where to place output files (temporarily)
        build_dir = pjoin(dirname(setup.app.doctreedir),
                          'nbplot_directive',
                          source_rel_dir)
        # get rid of .. in paths, also changes pathsep
        # see note in Python docs for warning about symbolic links on Windows.
        # need to compare source and dest paths at end
        build_dir = os.path.normpath(build_dir)

        if not exists(build_dir):
            os.makedirs(build_dir)

        # output_dir: final location in the builder's directory
        dest_dir = abspath(pjoin(setup.app.builder.outdir, source_rel_dir))
        if not exists(dest_dir):
            # no problem here for me, but just use built-ins
            os.makedirs(dest_dir)

        # We are now going to chose which parts of the text go to which
        # builders.  We might have specified which builders we want all the
        # text to go to, using the hide-from, and show-to options, or
        # (equivalently) using the include-source option. Or, we might have
        # split the code into parts, to send some of the code to the renderers,
        # such as HTML, and other parts of the code to the doctest builder
        # only.  For most builders, we do that by adding 'hide-from' and
        # 'show-to' attributes to the containing doctest node, and notifying
        # the builders visiting the node to discard the node, if those
        # attributes say they should.  However, the doctest builder does not
        # visit, so we have have to run some doctree post-processing on the
        # nbplot nodes, to check hide-from, show-to for the 'doctest' builder,
        # and, if the node should be hidden from the doctest builder, we
        # replace the doctest_block nodes with something that behaves in the
        # same way as a doctest_block node for all the visiting builders, but,
        # because it is not a doctest_block node, does not get picked up by the
        # doctest builder.

        # Break contents into parts, and select
        to_render, to_run = self._select_parts()

        # make figures
        try:
            images = render_figures(to_run,
                                    source_file_name,
                                    build_dir,
                                    output_base,
                                    config=config,
                                    context = True,  # keep plot context
                                    function_name = None,
                                    context_reset=context_reset,
                                    close_figs=close_figs,
                                    raises=raises)
            errors = []
        except PlotError as err:
            reporter = self.state.memo.reporter
            sm = reporter.system_message(
                2,
                'Exception plotting {output_base}\n'
                'from: {source_file_name}\n'
                'with code:\n\n{to_run}\n\n'
                'Exception:\n{err}'.format(**locals()),
                line=self.lineno)
            images = []
            errors = [sm]

        # generate output restructuredtext
        lines = [''] + [row.rstrip() for row in to_render.split('\n')]
        # If the code is not in doctest format, make it into code blocks.
        if not self._contains_doctest(to_render):
            lines = (['.. code-block:: python'] +
                     ['    ' + line for line in lines])
        # If we're going to run different code from the source we're rendering
        # then change node_attrs in place to hide the rendered code from the
        # doctest builder.
        if to_render != to_run:
            _hide_from_builder(node_attrs, 'doctest')
        # Build the rendered node
        rendered_nodes = self.rst2nodes(lines,
                                        self.nbplot_node,
                                        node_attrs)
        # Epilogue node contains the built figures and supporting stuff.
        epilogue = self._build_epilogue(images, source_rel_dir, build_dir)
        ret = rendered_nodes + epilogue + errors
        self._copy_image_files(images, dest_dir)
        # Now, we need to put in nodes for the code that ran, so the doctest
        # builder can find it.  But, we hide these nodes from all the other
        # builders (using hide-from all).
        if to_render != to_run and self._contains_doctest(to_run):
            lines = [''] + [row.rstrip() for row in to_run.split('\n')]
            ret += self.rst2nodes(
                lines,
                self.nbplot_node,
                {'hide-from': ['all'], 'show-to': ['doctest']})
        return ret


def _hide_from_builder(attrs, builder_name):
    """ Change ``show-to, hide-from`` values to hide from `builder_name`

    Modify `attrs` in-place.

    Parameters
    ----------
    attrs : dict
        Dict containing keys ``hide-from`` and ``show-to``.
    builder_name : str
        Name of builder that we want `attrs` to signal hiding from.
    """
    if builder_name in attrs['show-to']:
        attrs['show-to'].remove(builder_name)
    if builder_name not in attrs['hide-from']:
        attrs['hide-from'].append(builder_name)


#------------------------------------------------------------------------------
# Doctest handling
#------------------------------------------------------------------------------

def contains_doctest(text):
    try:
        # check if it's valid Python as-is
        compile(text, '<string>', 'exec')
        return False
    except SyntaxError:
        pass
    r = re.compile(r'^\s*>>>', re.M)
    m = r.search(text)
    return bool(m)


def unescape_doctest(text):
    """
    Extract code from a piece of text, which contains either Python code
    or doctests.

    """
    if not contains_doctest(text):
        return text

    code = ""
    for line in text.split("\n"):
        m = re.match(r'^\s*(>>>|\.\.\.) (.*)$', line)
        if m:
            code += m.group(2) + "\n"
        elif line.strip():
            code += "# " + line.strip() + "\n"
        else:
            code += "\n"
    return code


def remove_coding(text):
    """
    Remove the coding comment, which six.exec_ doesn't like.
    """
    sub_re = re.compile("^#\s*-\*-\s*coding:\s*.*-\*-$", flags=re.MULTILINE)
    return sub_re.sub("", text)

#------------------------------------------------------------------------------
# Template
#------------------------------------------------------------------------------


EPILOGUE_TEMPLATE = """
{{ only_html }}

   {% if html_show_formats and not multi_image %}
   (
   {%- for img in images -%}
     {%- for fmt in img.formats -%}
       {%- if not loop.first -%}, {% endif -%}
       `{{ fmt }} <{{ dest_dir }}/{{ img.basename }}.{{ fmt }}>`__
     {%- endfor -%}
   {%- endfor -%}
   )
   {% endif %}

   {% for img in images %}
   .. figure:: {{ build_dir }}/{{ img.basename }}.png
      {% for option in options -%}
      {{ option }}
      {% endfor %}

      {% if html_show_formats and multi_image -%}
        (
        {%- for fmt in img.formats -%}
        {%- if not loop.first -%}, {% endif -%}
        `{{ fmt }} <{{ dest_dir }}/{{ img.basename }}.{{ fmt }}>`__
        {%- endfor -%}
        )
      {%- endif -%}

   {% endfor %}

{{ only_latex }}

   {% for img in images %}
   {% if 'pdf' in img.formats -%}
   .. image:: {{ build_dir }}/{{ img.basename }}.pdf
   {% endif -%}
   {% endfor %}

{{ only_texinfo }}

   {% for img in images %}
   .. image:: {{ build_dir }}/{{ img.basename }}.png
      {% for option in options -%}
      {{ option }}
      {% endfor %}

   {% endfor %}

"""


# the context of the plot for all directives
plot_context = dict()


class ImageFile(object):
    def __init__(self, basename, path):
        self.basename = basename
        self.dirname = path
        self.formats = []

    def filename(self, format):
        return pjoin(self.dirname, "%s.%s" % (self.basename, format))

    def filenames(self):
        return [self.filename(fmt) for fmt in self.formats]


class PlotError(RuntimeError):
    pass


class PlotValueError(ValueError):
    pass


PARTER = re.compile(r"""(?:\n\n|^)\.\.\ part\n  # part separator
                    ((?:\ *\w+\ *=\ *.+\n)*)  # attributes
                    \n
                    """, re.VERBOSE)


ATTRIBUTER = re.compile(r'^(\w+) *= *(.+?) *$')


def _proc_part_def(part_def):
    """ Return part dictionary from `part_def` string
    """
    if part_def == '':
        return {}
    if not part_def.startswith(' '):
        raise PlotValueError('Part attributes should be indented')
    justified = textwrap.dedent(part_def).splitlines()
    if any(s.startswith(' ') for s in justified):
        raise PlotValueError('Part attributes should have same indentation')
    return dict(ATTRIBUTER.match(line).groups() for line in justified)


def _part_strs2dicts(part_strs):
    """ Return part def, contents pairs as part dictionaries
    """
    dicts = []
    while len(part_strs):
        part_dict = _proc_part_def(part_strs.pop(0))
        part_dict['contents'] = part_strs.pop(0).splitlines()
        dicts.append(part_dict)
    return dicts


def parse_parts(lines):
    """ Parse string list `content` into `parts`

    Parameters
    ----------
    lines : sequence of str
        Contents from directive.  Each element is a line.

    Returns
    -------
    parts : list
        List of dicts.  Each dict is a "part". Part dicts each have key
        ``contents``, with value being a list of strings, one string per line.
        The other key, value pairs are attributes of this part.
    """
    text = '\n'.join(lines).strip()
    part_strs = PARTER.split(text)
    if len(part_strs) == 1:
        return [{'contents': part_strs[0].splitlines()}]
    if part_strs[0] == '':  # begins with part attributes
        part_strs.pop(0)
    else:  # does not begin with part attributes, insert empty attributes
        part_strs.insert(0, '')
    return _part_strs2dicts(part_strs)


def _check_wd(path):
    try:
        abs_path = abspath(path)
    except TypeError as err:
        raise TypeError(str(err) + '\n`nbplot_working_directory` option in '
                        'Sphinx configuration file must be a string or '
                        'None')
    if not isdir(abs_path):
        raise OSError('`nbplot_working_directory` option (="{}") in '
                      'Sphinx configuration file must be a valid '
                      'directory path'.format(path))
    return abs_path


def run_code(code, code_path=None, ns=None, function_name=None, workdir=None,
             pre_code=None, raises=None):
    """
    Run `code` from file at `code_path` in namespace `ns`

    Parameters
    ----------
    code : str
        code to run.
    code_path : str
        Filename containing the code.
    ns : None or dict, optional
        Python namespace in which to execute code.  If None, make fresh
        namespace.
    function_name : None or str, optional
        If non-empty string, name of function to execute after executing
        `code`.
    workdir : None or str, optional
        Working directory in which to run code.  Defaults to current working
        directory.
    pre_code : None or str, optional
        Any code to run before `code`.
    raises : None or Exception class
        An exception that the run code should raise.

    Returns
    -------
    ns : dict
        Namespace, filled from execution of `code`.
    """
    # Change the working directory to the directory of the example, so
    # it can get at its data files, if any.  Add its path to sys.path
    # so it can import any helper modules sitting beside it.
    if six.PY2:
        pwd = os.getcwdu()
    else:
        pwd = os.getcwd()
    old_sys_path = list(sys.path)
    workdir = os.getcwd() if workdir is None else workdir
    os.chdir(workdir)
    sys.path.insert(0, workdir)

    # Reset sys.argv
    old_sys_argv = sys.argv
    sys.argv = [code_path]

    # Redirect stdout
    stdout = sys.stdout
    sys.stdout = io.StringIO() if six.PY3 else io.BytesIO()

    # Assign a do-nothing print function to the namespace.  There
    # doesn't seem to be any other way to provide a way to (not) print
    # that works correctly across Python 2 and 3.
    def _dummy_print(*arg, **kwarg):
        pass

    ns = {} if ns is None else ns
    try:
        try:
            code = unescape_doctest(code)
            if pre_code and not ns:
                six.exec_(six.text_type(pre_code), ns)
            ns['print'] = _dummy_print
            if "__main__" in code:
                six.exec_("__name__ = '__main__'", ns)
            code = remove_coding(code)
            if raises is None:
                six.exec_(code, ns)
            else:  # Code should raise exception
                try:
                    six.exec_(code, ns)
                except raises:
                    pass
            if function_name:
                six.exec_(function_name + "()", ns)
        except (Exception, SystemExit):
            raise PlotError(traceback.format_exc())
    finally:
        os.chdir(pwd)
        sys.argv = old_sys_argv
        sys.path[:] = old_sys_path
        sys.stdout = stdout
    return ns


def render_figures(code, code_path, output_dir, output_base, config,
                   context=True, function_name=None, context_reset=False,
                   close_figs=False, raises=None):
    """ Run plot code and save the hi/low res PNGs, PDF in `output_dir`

    Save the images under `output_dir` with file names derived from
    `output_base`.

    Parameters
    ----------
    code : str
        String containing code to run.
    code_path : str
        Path of file containing code.  Usually path to ``.rst`` file.
    output_dir : str
        Path to which to write output images from plots.
    output_base : str
        Prefix for filename(s) for output image(s).
    config : instance
        Sphinx configuration instance.
    context : {True, False}, optional
        If True, use persistent context (workspace) for executing code.
        Otherwise create new empty context for executing code.
    function_name : None or str, optional
        If not-empty str, name of function to execute after executing `code`.
    context_reset : {False, True}, optional
        If True, clear persistent context (workspace) for code.
    close_figs : {False, True}, optional
        If True, close all figures generated before our `code` runs.  False can
        be useful when building up a plot with several `code` blocks.
    raises : None or Exception, optional
        Exception class that code should raise, or None, for no exception.
    """
    # -- Parse format list
    default_dpi = {'png': 80, 'hires.png': 200, 'pdf': 200}
    formats = []
    plot_formats = config.nbplot_formats
    if isinstance(plot_formats, six.string_types):
        # String Sphinx < 1.3, Split on , to mimic
        # Sphinx 1.3 and later. Sphinx 1.3 always
        # returns a list.
        plot_formats = plot_formats.split(',')
    for fmt in plot_formats:
        if isinstance(fmt, six.string_types):
            if ':' in fmt:
                suffix,dpi = fmt.split(':')
                formats.append((str(suffix), int(dpi)))
            else:
                formats.append((fmt, default_dpi.get(fmt, 80)))
        elif type(fmt) in (tuple, list) and len(fmt)==2:
            formats.append((str(fmt[0]), int(fmt[1])))
        else:
            raise PlotError('invalid image format "%r" in nbplot_formats' % fmt)

    # Build the output
    ns = plot_context if context else {}

    if context_reset:
        plt.close('all')
        matplotlib.rc_file_defaults()
        matplotlib.rcParams.update(config.nbplot_rcparams)
        plot_context.clear()

    close_figs = not context or close_figs

    # Get working directory for code execution
    if setup.config.nbplot_working_directory is not None:
        workdir = _check_wd(setup.config.nbplot_working_directory)
    elif code_path is not None:
        workdir = abspath(dirname(code_path))
    else:
        workdir = None

    if close_figs:
        plt.close('all')

    run_code(code, code_path, ns, function_name, workdir=workdir,
             pre_code=setup.config.nbplot_pre_code, raises=raises)

    images = []
    fig_managers = Gcf.get_all_fig_managers()
    for j, figman in enumerate(fig_managers):
        if len(fig_managers) == 1:
            img = ImageFile(output_base, output_dir)
        else:
            img = ImageFile("%s_%02d" % (output_base, j), output_dir)
        images.append(img)
        for format, dpi in formats:
            try:
                figman.canvas.figure.savefig(img.filename(format), dpi=dpi)
            except Exception:
                raise PlotError(traceback.format_exc())
            img.formats.append(format)

    return images


# Sphinx event handlers

def _false():
    # Must be function rather than lambda to allow pickling of environment
    return False


def do_builder_init(app):
    env = app.env
    env.nbplot_reset_markers = defaultdict(_false)
    env.nbplot_flag_namespaces = defaultdict(dict)


def do_purge_doc(app, env, docname):
    """ Clear markers for whether `docname` has seen a plot context reset
    """
    env.nbplot_reset_markers[docname] = False
    env.nbplot_flag_namespaces[docname] = env.config.nbplot_flags.copy()


def likes_builder(node, builder_name):
    return (not hasattr(node, 'likes_builder') or
            node.likes_builder(builder_name))


def checked_visit(self, node):
    if not likes_builder(node, self.builder.name):
        raise nodes.SkipNode


def checked_depart(self, node):
    pass


def dont_doctest_visit(self, node):
    # Make node behave like doctest node, for builders that visit.  This allows
    # us to make another not-doctest_block node to replace the actual
    # ductest_block node, so that, for builders that visit, such as html, we
    # get the same output, but thd doctest builder passes over it, because it
    # is not a doctest_block node.
    self.visit_doctest_block(node)


def dont_doctest_depart(self, node):
    self.depart_doctest_block(node)


def setup(app):
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    # Builders which run visit methods on nodes.  Basically everything but
    # doctest.
    visiting_builders = ('html', 'latex', 'text', 'texinfo')

    # Containers used as markers for nbplot contents
    for node_class in (nbplot_container, nbplot_epilogue):
        app.add_node(node_class,
                     **{builder: (checked_visit, checked_depart)
                        for builder in visiting_builders})
    # Allow nodes hidden from doctest builder to render on other builders
    app.add_node(dont_doctest_block,
                 **{builder: (dont_doctest_visit, dont_doctest_depart)
                    for builder in visiting_builders})
    app.add_directive('nbplot', NBPlotDirective)
    app.add_directive('nbplot-flags', NBPlotFlags)
    app.add_directive('nbplot-show-flags', NBPlotShowFlags)
    pre_default = "import numpy as np\nfrom matplotlib import pyplot as plt\n"
    app.add_config_value('nbplot_pre_code', pre_default, True)
    app.add_config_value('nbplot_include_source', True, True)
    app.add_config_value('nbplot_formats', ['png', 'hires.png', 'pdf'], True)
    app.add_config_value('nbplot_html_show_formats', True, True)
    app.add_config_value('nbplot_rcparams', {}, True)
    app.add_config_value('nbplot_working_directory', None, True)
    app.add_config_value('nbplot_template', None, True)
    app.add_config_value('nbplot_flags', {}, True)

    # Create dictionaries in builder environment
    app.connect(str('builder-inited'), do_builder_init)
    # Clear marker indicating that we have already started parsing a page
    app.connect(str('env-purge-doc'), do_purge_doc)



""" Test that timeout parameters work as expected
"""

from os.path import join as pjoin, isfile

from nb2plots.testing import PlotsBuilder


class _CheckTimeout(PlotsBuilder):
    """ Machinery for checking timeout
    """

    rst_sources = {'a_page': r"""\
Title
#####

.. code-links::

.. nbplot::

    >>> from time import sleep
    >>> sleep(2)
"""}


class TestNoTimeout(_CheckTimeout):
    """ The default timeout is fairly long, so a short wait is OK

    ``ExecutePreprocessor`` in ``nbconvert/preprocessors/execute.py`` suggests
    default is 30 seconds.
    """

    def test_output(self):
        assert isfile(pjoin(self.out_dir, 'a_page_full.ipynb'))


class TestCLOptsTimeout(_CheckTimeout):
    """ Set timeout to 1 second with code-link option, build should error
    """

    rst_sources = {'a_page': r"""\
Title
#####

.. code-links::
    :timeout: 1

.. nbplot::

    >>> from time import sleep
    >>> sleep(2)
"""}

    should_error = True


class TestFTimeout(_CheckTimeout):
    """ Set timeout to 1 second with role option, build should error
    """

    rst_sources = {'a_page': r"""\
Title
#####

.. role:: longerfullnotebook(fullnotebook)
   :timeout: 1

Add :longerfullnotebook:`.`

.. nbplot::

    >>> from time import sleep
    >>> sleep(2)
"""}

    should_error = True


class TestConfTimeout(_CheckTimeout):
    """ Set timeout to 1 second with conf.py option, build should error
    """

    conf_source = (TestNoTimeout.conf_source +
                   '\nfill_notebook_timeout = 1')

    should_error = True


class TestConfigCLTimeout(TestNoTimeout):
    """ Code-links options overrides configuration option
    """

    conf_source = (TestNoTimeout.conf_source +
                   '\nfill_notebook_timeout = 1')

    rst_sources = {'a_page': r"""\
Title
#####

.. code-links::
    :timeout: 10

.. nbplot::

    >>> from time import sleep
    >>> sleep(2)
"""}


class TestBestTimeout(TestNoTimeout):
    """ Check that multiple links to same notebook use longest timeout
    """

    rst_sources = {'a_page': r"""\
Title
#####

.. code-links::
    :timeout: 1

.. role:: shorterfullnotebook(fullnotebook)
   :timeout: 2

Add :shorterfullnotebook:`short <short.ipynb>`

.. role:: longerfullnotebook(fullnotebook)
   :timeout: 10

Add :longerfullnotebook:`long <long.ipynb>`

.. nbplot::

    >>> from time import sleep
    >>> sleep(3)
"""}



# Here are some:

# doctest blocks
a = 1
a

# Interspersed with text.

# another block
b = [1, 2, 3]
b

# Explaining here.

# Here is a comment line that goes beyond the normal 80 character line wrap to test (non) wrapping



#!/usr/bin/env python
# -*- coding: utf-8 -*-
''' Installation script for nb2plots package '''
import os
from os.path import join as pjoin, split as psplit, splitext
import re

import setuptools
from distutils.core import setup

import versioneer

# Get install requirements from requirements.txt file
with open('requirements.txt', 'rt') as fobj:
    install_requires = [line.strip() for line in fobj
                        if line.strip() and not line[0] in '#-']
# Get any extra test requirements
with open('test-requirements.txt', 'rt') as fobj:
    test_requires = [line.strip() for line in fobj
                     if line.strip() and not line[0] in '#-']

# Requires for distutils (only used in pypi interface?)
break_ver = re.compile(r'(\S+?)(\[\S+\])?([=<>!]+\S+)')
requires = [break_ver.sub(r'\1 (\3)', req) for req in install_requires]

# See: https://github.com/matthew-brett/myscripter
from distutils.command.install_scripts import install_scripts
from distutils import log

BAT_TEMPLATE = \
r"""@echo off
REM wrapper to use shebang first line of {FNAME}
set mypath=%~dp0
set pyscript="%mypath%{FNAME}"
set /p line1=<%pyscript%
if "%line1:~0,2%" == "#!" (goto :goodstart)
echo First line of %pyscript% does not start with "#!"
exit /b 1
:goodstart
set py_exe=%line1:~2%
call "%py_exe%" %pyscript% %*
"""

class my_install_scripts(install_scripts):
    """ Install .bat wrapper for scripts on Windows """
    def run(self):
        install_scripts.run(self)
        if not os.name == "nt":
            return
        for filepath in self.get_outputs():
            # If we can find an executable name in the #! top line of the
            # script file, make .bat wrapper for script.
            with open(filepath, 'rt') as fobj:
                first_line = fobj.readline()
            if not (first_line.startswith('#!') and
                    'python' in first_line.lower()):
                log.info("No #!python executable found, skipping .bat "
                            "wrapper")
                continue
            pth, fname = psplit(filepath)
            froot, ext = splitext(fname)
            bat_file = pjoin(pth, froot + '.bat')
            bat_contents = BAT_TEMPLATE.replace('{FNAME}', fname)
            log.info("Making %s wrapper for %s" % (bat_file, filepath))
            if self.dry_run:
                continue
            with open(bat_file, 'wt') as fobj:
                fobj.write(bat_contents)


cmdclass = versioneer.get_cmdclass()
cmdclass['install_scripts'] = my_install_scripts


setup(name='nb2plots',
      version=versioneer.get_version(),
      cmdclass=cmdclass,
      description='Converting between ipython notebooks and sphinx docs',
      author='Matthew Brett',
      author_email='matthew.brett@gmail.com',
      maintainer='Matthew Brett',
      maintainer_email='matthew.brett@gmail.com',
      url='http://github.com/matthew-brett/nb2plots',
      packages=['nb2plots',
                'nb2plots.tests',
                'nb2plots.testing'],
      package_data = {'nb2plots': [
          'tests/data/*.rst',
          'tests/rst_md_files/*',
          'tests/data/*',
          'tests/proj1/*.rst',
          'tests/proj1/*.py',
          'tests/proj1/_static/*',
          'tests/otherpages/*',
      ]},
      license='BSD license',
      classifiers = [
          'Development Status :: 2 - Pre-Alpha',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Topic :: Scientific/Engineering',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Operating System :: Unix',
          'Operating System :: MacOS',
        ],
      scripts = ['scripts/nb2plots',
                 'scripts/sphinx2nb',
                 'scripts/sphinx2py',
                 'scripts/sphinx2md',
                 'scripts/sphinx2pxml',
                 'scripts/rst2md'],
      long_description = open('README.rst', 'rt').read(),
      install_requires = install_requires,
      extras_require = {'test': test_requires}
      )
